# type: ignore
import logging
import os
import time
import warnings
from enum import Enum
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)

from labelbox.orm.db_object import DbObject, experimental
from labelbox.orm.model import Entity, Field, Relationship
from labelbox.orm.query import results_query_part
from labelbox.pagination import PaginatedCollection
from labelbox.schema.conflict_resolution_strategy import (
    ConflictResolutionStrategy,
)
from labelbox.schema.export_params import ModelRunExportParams
from labelbox.schema.export_task import ExportTask
from labelbox.schema.identifiables import DataRowIds, GlobalKeys
from labelbox.schema.send_to_annotate_params import (
    SendToAnnotateFromModelParams,
    build_destination_task_queue_input,
    build_predictions_input,
)
from labelbox.schema.task import Task

if TYPE_CHECKING:
    from labelbox import MEAPredictionImport
    from labelbox.types import Label

logger = logging.getLogger(__name__)

DATAROWS_IMPORT_LIMIT = 25000


class DataSplit(Enum):
    TRAINING = "TRAINING"
    TEST = "TEST"
    VALIDATION = "VALIDATION"
    UNASSIGNED = "UNASSIGNED"


class ModelRun(DbObject):
    name = Field.String("name")
    updated_at = Field.DateTime("updated_at")
    created_at = Field.DateTime("created_at")
    created_by_id = Field.String("created_by_id", "createdBy")
    model_id = Field.String("model_id")
    training_metadata = Field.Json("training_metadata")

    class Status(Enum):
        EXPORTING_DATA = "EXPORTING_DATA"
        PREPARING_DATA = "PREPARING_DATA"
        TRAINING_MODEL = "TRAINING_MODEL"
        COMPLETE = "COMPLETE"
        FAILED = "FAILED"

    def upsert_labels(
        self,
        label_ids: Optional[List[str]] = None,
        project_id: Optional[str] = None,
        timeout_seconds=3600,
    ):
        """
        Adds data rows and labels to a Model Run

        Args:
            label_ids (list): label ids to insert
            project_id (string): project uuid, all project labels will be uploaded
                Either label_ids OR project_id is required but NOT both
            timeout_seconds (float): Max waiting time, in seconds.
        Returns:
            ID of newly generated async task

        """

        use_label_ids = label_ids is not None and len(label_ids) > 0
        use_project_id = project_id is not None

        if not use_label_ids and not use_project_id:
            raise ValueError(
                "Must provide at least one label id or a project id"
            )

        if use_label_ids and use_project_id:
            raise ValueError("Must only one of label ids, project id")

        if use_label_ids:
            return self._upsert_labels_by_label_ids(label_ids, timeout_seconds)
        else:  # use_project_id
            return self._upsert_labels_by_project_id(
                project_id, timeout_seconds
            )

    def _upsert_labels_by_label_ids(
        self, label_ids: List[str], timeout_seconds: int
    ):
        mutation_name = "createMEAModelRunLabelRegistrationTask"
        create_task_query_str = """mutation createMEAModelRunLabelRegistrationTaskPyApi($modelRunId: ID!, $labelIds : [ID!]!) {
        %s(where : { id : $modelRunId}, data : {labelIds: $labelIds})}
        """ % (mutation_name)

        res = self.client.execute(
            create_task_query_str,
            {"modelRunId": self.uid, "labelIds": label_ids},
        )
        task_id = res[mutation_name]

        status_query_str = """query MEALabelRegistrationTaskStatusPyApi($where: WhereUniqueIdInput!){
            MEALabelRegistrationTaskStatus(where: $where) {status errorMessage}
        }
        """
        return self._wait_till_done(
            lambda: self.client.execute(
                status_query_str, {"where": {"id": task_id}}
            )["MEALabelRegistrationTaskStatus"],
            timeout_seconds=timeout_seconds,
        )

    def _upsert_labels_by_project_id(
        self, project_id: str, timeout_seconds: int
    ):
        mutation_name = "createMEAModelRunProjectLabelRegistrationTask"
        create_task_query_str = """mutation createMEAModelRunProjectLabelRegistrationTaskPyApi($modelRunId: ID!, $projectId : ID!) {
        %s(where : { modelRunId : $modelRunId, projectId: $projectId})}
        """ % (mutation_name)

        res = self.client.execute(
            create_task_query_str,
            {"modelRunId": self.uid, "projectId": project_id},
        )
        task_id = res[mutation_name]

        status_query_str = """query MEALabelRegistrationTaskStatusPyApi($where: WhereUniqueIdInput!){
            MEALabelRegistrationTaskStatus(where: $where) {status errorMessage}
        }
        """
        return self._wait_till_done(
            lambda: self.client.execute(
                status_query_str, {"where": {"id": task_id}}
            )["MEALabelRegistrationTaskStatus"],
            timeout_seconds=timeout_seconds,
        )

    def upsert_data_rows(
        self, data_row_ids=None, global_keys=None, timeout_seconds=3600
    ):
        """Adds data rows to a Model Run without any associated labels
        Args:
            data_row_ids (list): data row ids to add to model run
            global_keys (list): global keys for data rows to add to model run
            timeout_seconds (float): Max waiting time, in seconds.
        Returns:
            ID of newly generated async task
        """

        mutation_name = "createMEAModelRunDataRowRegistrationTask"
        create_task_query_str = """mutation createMEAModelRunDataRowRegistrationTaskPyApi($modelRunId: ID!, $dataRowIds: [ID!], $globalKeys: [ID!]) {
          %s(where : { id : $modelRunId}, data : {dataRowIds: $dataRowIds, globalKeys: $globalKeys})}
          """ % (mutation_name)

        res = self.client.execute(
            create_task_query_str,
            {
                "modelRunId": self.uid,
                "dataRowIds": data_row_ids,
                "globalKeys": global_keys,
            },
        )
        task_id = res[mutation_name]

        status_query_str = """query MEADataRowRegistrationTaskStatusPyApi($where: WhereUniqueIdInput!){
            MEADataRowRegistrationTaskStatus(where: $where) {status errorMessage}
        }
        """
        return self._wait_till_done(
            lambda: self.client.execute(
                status_query_str, {"where": {"id": task_id}}
            )["MEADataRowRegistrationTaskStatus"],
            timeout_seconds=timeout_seconds,
        )

    def _wait_till_done(self, status_fn, timeout_seconds=120, sleep_time=5):
        # Do not use this function outside of the scope of upsert_data_rows or upsert_labels. It could change.
        original_timeout = timeout_seconds
        while True:
            res = status_fn()
            if res["status"] == "COMPLETE":
                return True
            elif res["status"] == "FAILED":
                raise Exception(f"Job failed.")
            timeout_seconds -= sleep_time
            if timeout_seconds <= 0:
                raise TimeoutError(
                    f"Unable to complete import within {original_timeout} seconds."
                )
            time.sleep(sleep_time)

    def upsert_predictions_and_send_to_project(
        self,
        name: str,
        predictions: Union[str, Path, Iterable[Dict]],
        project_id: str,
        priority: Optional[int] = 5,
    ) -> "MEAPredictionImport":  # type: ignore
        """
        Provides a convenient way to execute the following steps in a single function call:
            1. Upload predictions to a Model
            2. Create a batch from data rows that had predictions assocated with them
            3. Attach the batch to a project
            4. Add those same predictions to the project as MAL annotations

        Note that partial successes are possible.
        If it is important that all stages are successful then check the status of each individual task
        with task.errors. E.g.

        >>>    mea_import_job, batch, mal_import_job = upsert_predictions_and_send_to_project(name, predictions, project_id)
        >>>    # handle mea import job successfully created (check for job failure or partial failures)
        >>>    print(mea_import_job.status, mea_import_job.errors)
        >>>    if batch is None:
        >>>        # Handle batch creation failure
        >>>    if mal_import_job is None:
        >>>        # Handle mal_import_job creation failure
        >>>    else:
        >>>        # handle mal import job successfully created (check for job failure or partial failures)
        >>>        print(mal_import_job.status, mal_import_job.errors)


        Args:
            name (str): name of the AnnotationImport job as well as the name of the batch import
            predictions (Iterable):
                iterable of annotation rows
            project_id (str): id of the project to import into
            priority (int): priority of the job
        Returns:
            Tuple[MEAPredictionImport, Batch, MEAToMALPredictionImport]
            If any of these steps fail the return value will be None.

        """
        kwargs = dict(client=self.client, model_run_id=self.uid, name=name)
        project = self.client.get_project(project_id)
        import_job = self.add_predictions(name, predictions)
        prediction_statuses = import_job.statuses
        mea_to_mal_data_rows = list(
            set(
                [
                    row["dataRow"]["id"]
                    for row in prediction_statuses
                    if row["status"] == "SUCCESS"
                ]
            )
        )

        if not mea_to_mal_data_rows:
            # 0 successful model predictions imported
            return import_job, None, None

        elif len(mea_to_mal_data_rows) >= DATAROWS_IMPORT_LIMIT:
            mea_to_mal_data_rows = mea_to_mal_data_rows[:DATAROWS_IMPORT_LIMIT]
            logger.warning(
                f"Exeeded max data row limit {len(mea_to_mal_data_rows)}, trimmed down to {DATAROWS_IMPORT_LIMIT} data rows."
            )

        try:
            batch = project.create_batch(name, mea_to_mal_data_rows, priority)
        except Exception as e:
            logger.warning(f"Failed to create batch. Messsage : {e}.")
            # Unable to create batch
            return import_job, None, None

        try:
            mal_prediction_import = (
                Entity.MEAToMALPredictionImport.create_for_model_run_data_rows(
                    data_row_ids=mea_to_mal_data_rows,
                    project_id=project_id,
                    **kwargs,
                )
            )
            mal_prediction_import.wait_until_done()
        except Exception as e:
            logger.warning(
                f"Failed to create MEA to MAL prediction import. Message : {e}."
            )
            # Unable to create mea to mal prediction import
            return import_job, batch, None

        return import_job, batch, mal_prediction_import

    def add_predictions(
        self,
        name: str,
        predictions: Union[str, Path, Iterable[Dict], Iterable["Label"]],
    ) -> "MEAPredictionImport":  # type: ignore
        """
        Uploads predictions to a new Editor project.

        Args:
            name (str): name of the AnnotationImport job
            predictions (str or Path or Iterable): url that is publicly accessible by Labelbox containing an
                ndjson file
                OR local path to an ndjson file
                OR iterable of annotation rows

        Returns:
            AnnotationImport
        """
        kwargs = dict(client=self.client, id=self.uid, name=name)
        if isinstance(predictions, str) or isinstance(predictions, Path):
            if os.path.exists(predictions):
                return Entity.MEAPredictionImport.create(
                    path=str(predictions), **kwargs
                )
            else:
                return Entity.MEAPredictionImport.create(
                    url=str(predictions), **kwargs
                )
        elif isinstance(predictions, Iterable):
            return Entity.MEAPredictionImport.create(
                labels=predictions, **kwargs
            )
        else:
            raise ValueError(
                f"Invalid predictions given of type: {type(predictions)}"
            )

    def model_run_data_rows(self):
        query_str = """query modelRunPyApi($modelRunId: ID!, $from : String, $first: Int){
                annotationGroups(where: {modelRunId: {id: $modelRunId}}, after: $from, first: $first)
                {nodes{%s},pageInfo{endCursor}}
            }
        """ % (results_query_part(ModelRunDataRow))
        return PaginatedCollection(
            self.client,
            query_str,
            {"modelRunId": self.uid},
            ["annotationGroups", "nodes"],
            lambda client, res: ModelRunDataRow(client, self.model_id, res),
            ["annotationGroups", "pageInfo", "endCursor"],
        )

    def delete(self):
        """Deletes specified Model Run.

        Returns:
            Query execution success.
        """
        ids_param = "ids"
        query_str = """mutation DeleteModelRunPyApi($%s: ID!) {
            deleteModelRuns(where: {ids: [$%s]})}""" % (ids_param, ids_param)
        self.client.execute(query_str, {ids_param: str(self.uid)})

    def delete_model_run_data_rows(self, data_row_ids: List[str]):
        """Deletes data rows from Model Runs.

        Args:
            data_row_ids (list): List of data row ids to delete from the Model Run.
        Returns:
            Query execution success.
        """
        model_run_id_param = "modelRunId"
        data_row_ids_param = "dataRowIds"
        query_str = """mutation DeleteModelRunDataRowsPyApi($%s: ID!, $%s: [ID!]!) {
            deleteModelRunDataRows(where: {modelRunId: $%s, dataRowIds: $%s})}""" % (
            model_run_id_param,
            data_row_ids_param,
            model_run_id_param,
            data_row_ids_param,
        )
        self.client.execute(
            query_str,
            {model_run_id_param: self.uid, data_row_ids_param: data_row_ids},
        )

    @experimental
    def assign_data_rows_to_split(
        self,
        data_row_ids: List[str] = None,
        split: Union[DataSplit, str] = None,
        global_keys: List[str] = None,
        timeout_seconds=120,
    ):
        split_value = split.value if isinstance(split, DataSplit) else split
        valid_splits = DataSplit._member_names_

        if split_value is None or split_value not in valid_splits:
            raise ValueError(
                f"`split` must be one of : `{valid_splits}`. Found : `{split}`"
            )

        task_id = self.client.execute(
            """mutation assignDataSplitPyApi($modelRunId: ID!, $data: CreateAssignDataRowsToDataSplitTaskInput!){
                  createAssignDataRowsToDataSplitTask(modelRun : {id: $modelRunId}, data: $data)}
            """,
            {
                "modelRunId": self.uid,
                "data": {
                    "assignments": [
                        {
                            "split": split_value,
                            "dataRowIds": data_row_ids,
                            "globalKeys": global_keys,
                        }
                    ]
                },
            },
            experimental=True,
        )["createAssignDataRowsToDataSplitTask"]

        status_query_str = """query assignDataRowsToDataSplitTaskStatusPyApi($id: ID!){
            assignDataRowsToDataSplitTaskStatus(where: {id : $id}){status errorMessage}}
            """

        return self._wait_till_done(
            lambda: self.client.execute(
                status_query_str, {"id": task_id}, experimental=True
            )["assignDataRowsToDataSplitTaskStatus"],
            timeout_seconds=timeout_seconds,
        )

    @experimental
    def update_status(
        self,
        status: Union[str, "ModelRun.Status"],
        metadata: Optional[Dict[str, str]] = None,
        error_message: Optional[str] = None,
    ):
        status_value = (
            status.value if isinstance(status, ModelRun.Status) else status
        )
        if status_value not in ModelRun.Status._member_names_:
            raise ValueError(
                f"Status must be one of : `{ModelRun.Status._member_names_}`. Found : `{status_value}`"
            )

        data: Dict[str, Any] = {"status": status_value}
        if error_message:
            data["errorMessage"] = error_message

        if metadata:
            data["metadata"] = metadata

        self.client.execute(
            """mutation setPipelineStatusPyApi($modelRunId: ID!, $data: UpdateTrainingPipelineInput!){
                updateTrainingPipeline(modelRun: {id : $modelRunId}, data: $data){status}
            }
        """,
            {"modelRunId": self.uid, "data": data},
            experimental=True,
        )

    def update_config(self, config: Dict[str, Any]) -> Dict[str, Any]:
        """
        Updates the Model Run's training metadata config
        Args:
            config (dict): A dictionary of keys and values
        Returns:
            Model Run id and updated training metadata
        """
        data: Dict[str, Any] = {"config": config}
        res = self.client.execute(
            """mutation updateModelRunConfigPyApi($modelRunId: ID!, $data: UpdateModelRunConfigInput!){
                updateModelRunConfig(modelRun: {id : $modelRunId}, data: $data){trainingMetadata}
            }
        """,
            {"modelRunId": self.uid, "data": data},
        )
        return res["updateModelRunConfig"]

    def reset_config(self) -> Dict[str, Any]:
        """
        Resets Model Run's training metadata config
        Returns:
            Model Run id and reset training metadata
        """
        res = self.client.execute(
            """mutation resetModelRunConfigPyApi($modelRunId: ID!){
                resetModelRunConfig(modelRun: {id : $modelRunId}){trainingMetadata}
            }
        """,
            {"modelRunId": self.uid},
        )
        return res["resetModelRunConfig"]

    def get_config(self) -> Dict[str, Any]:
        """
        Gets Model Run's training metadata
        Returns:
            training metadata as a dictionary
        """
        res = self.client.execute(
            """query ModelRunPyApi($modelRunId: ID!){
                modelRun(where: {id : $modelRunId}){trainingMetadata}
            }
        """,
            {"modelRunId": self.uid},
        )
        return res["modelRun"]["trainingMetadata"]

    def export(
        self,
        task_name: Optional[str] = None,
        params: Optional[ModelRunExportParams] = None,
    ) -> ExportTask:
        """
        Creates a model run export task with the given params and returns the task.

        >>>    export_task = export("my_export_task", params={"media_attributes": True})

        """
        task, _ = self._export(task_name, params, streamable=True)
        return ExportTask(task)

    def export_v2(
        self,
        task_name: Optional[str] = None,
        params: Optional[ModelRunExportParams] = None,
    ) -> Union[Task, ExportTask]:
        """
        Creates a model run export task with the given params and returns the task.

        >>>    export_task = export_v2("my_export_task", params={"media_attributes": True})

        """

        warnings.warn(
            "The method export_v2 for ModelRun is deprecated and will be removed in the next major release. Use the export method instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        task, is_streamable = self._export(task_name, params)
        if is_streamable:
            return ExportTask(task, True)
        return task

    def _export(
        self,
        task_name: Optional[str] = None,
        params: Optional[ModelRunExportParams] = None,
        streamable: bool = False,
    ) -> Tuple[Task, bool]:
        mutation_name = "exportDataRowsInModelRun"
        create_task_query_str = (
            f"mutation {mutation_name}PyApi"
            f"($input: ExportDataRowsInModelRunInput!)"
            f"{{{mutation_name}(input: $input){{taskId isStreamable}}}}"
        )

        _params = params or ModelRunExportParams()

        query_params = {
            "input": {
                "taskName": task_name,
                "filters": {"modelRunId": self.uid},
                "isStreamableReady": True,
                "params": {
                    "mediaTypeOverride": _params.get(
                        "media_type_override", None
                    ),
                    "includeAttachments": _params.get("attachments", False),
                    "includeEmbeddings": _params.get("embeddings", False),
                    "includeMetadata": _params.get("metadata_fields", False),
                    "includeDataRowDetails": _params.get(
                        "data_row_details", False
                    ),
                    "includePredictions": _params.get("predictions", False),
                    "includeModelRunDetails": _params.get(
                        "model_run_details", False
                    ),
                },
                "streamable": streamable,
            }
        }
        res = self.client.execute(
            create_task_query_str, query_params, error_log_key="errors"
        )
        res = res[mutation_name]
        task_id = res["taskId"]
        is_streamable = res["isStreamable"]
        return Task.get_task(self.client, task_id), is_streamable

    def send_to_annotate_from_model(
        self,
        destination_project_id: str,
        task_queue_id: Optional[str],
        batch_name: str,
        data_rows: Union[DataRowIds, GlobalKeys],
        params: SendToAnnotateFromModelParams,
    ) -> Task:
        """
        Sends data rows from a model run to a project for annotation.

        Example Usage:
            >>> task = model_run.send_to_annotate_from_model(
            >>>     destination_project_id=DESTINATION_PROJECT_ID,
            >>>     batch_name="batch",
            >>>     data_rows=UniqueIds([DATA_ROW_ID]),
            >>>     task_queue_id=TASK_QUEUE_ID,
            >>>     params={})
            >>> task.wait_till_done()

        Args:
            destination_project_id: The ID of the project to send the data rows to.
            task_queue_id: The ID of the task queue to send the data rows to.  If not specified, the data rows will be
                sent to the Done workflow state.
            batch_name: The name of the batch to create. If more than one batch is created, additional batches will be
                named with a monotonically increasing numerical suffix, starting at "_1".
            data_rows: The data rows to send to the project.
            params: Additional parameters for this operation. See SendToAnnotateFromModelParams for details.

        Returns: The created task for this operation.

        """

        mutation_str = """mutation SendToAnnotateFromMeaPyApi($input: SendToAnnotateFromMeaInput!) {
                            sendToAnnotateFromMea(input: $input) {
                              taskId
                            }
                          }
        """

        destination_task_queue = build_destination_task_queue_input(
            task_queue_id
        )
        data_rows_query = self.client.build_catalog_query(data_rows)

        predictions_ontology_mapping = params.get(
            "predictions_ontology_mapping", None
        )
        predictions_input = build_predictions_input(
            predictions_ontology_mapping, self.uid
        )

        batch_priority = params.get("batch_priority", 5)
        exclude_data_rows_in_project = params.get(
            "exclude_data_rows_in_project", False
        )
        override_existing_annotations_rule = params.get(
            "override_existing_annotations_rule",
            ConflictResolutionStrategy.KeepExisting,
        )
        res = self.client.execute(
            mutation_str,
            {
                "input": {
                    "destinationProjectId": destination_project_id,
                    "batchInput": {
                        "batchName": batch_name,
                        "batchPriority": batch_priority,
                    },
                    "destinationTaskQueue": destination_task_queue,
                    "excludeDataRowsInProject": exclude_data_rows_in_project,
                    "annotationsInput": None,
                    "predictionsInput": predictions_input,
                    "conflictLabelsResolutionStrategy": override_existing_annotations_rule,
                    "searchQuery": [data_rows_query],
                    "sourceModelRunId": self.uid,
                }
            },
        )["sendToAnnotateFromMea"]

        return Entity.Task.get_task(self.client, res["taskId"])


class ModelRunDataRow(DbObject):
    label_id = Field.String("label_id")
    model_run_id = Field.String("model_run_id")
    data_split = Field.Enum(DataSplit, "data_split")
    data_row = Relationship.ToOne("DataRow", False, cache=True)

    def __init__(self, client, model_id, *args, **kwargs):
        super().__init__(client, *args, **kwargs)
        self.model_id = model_id

    @property
    def url(self):
        app_url = self.client.app_url
        endpoint = f"{app_url}/models/{self.model_id}/{self.model_run_id}/AllDatarowsSlice/{self.uid}?view=carousel"
        return endpoint



# type: ignore
from copy import deepcopy
from datetime import datetime
from enum import Enum
from itertools import chain
from typing import (
    Annotated,
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    Type,
    Union,
    overload,
)

from pydantic import (
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    StringConstraints,
    conlist,
    model_serializer,
)
from typing_extensions import Annotated

from labelbox.schema.identifiable import GlobalKey, UniqueId
from labelbox.schema.identifiables import DataRowIdentifiers, UniqueIds
from labelbox.schema.tool_building.types import SchemaId
from labelbox.utils import (
    _CamelCaseMixin,
    format_iso_datetime,
    format_iso_from_string,
)

Name = Annotated[
    str,
    BeforeValidator(lambda x: str.strip(str(x))),
    Field(min_length=1, max_length=100),
]


class DataRowMetadataKind(Enum):
    number = "CustomMetadataNumber"
    datetime = "CustomMetadataDateTime"
    enum = "CustomMetadataEnum"
    string = "CustomMetadataString"
    option = "CustomMetadataEnumOption"
    embedding = "CustomMetadataEmbedding"


# Metadata schema
class DataRowMetadataSchema(BaseModel):
    uid: SchemaId
    name: Name
    reserved: bool
    kind: DataRowMetadataKind
    options: Optional[List["DataRowMetadataSchema"]] = None
    parent: Optional[SchemaId] = None


DataRowMetadataSchema.model_rebuild()

Embedding: Type[List[float]] = conlist(float, min_length=128, max_length=128)
String: Type[str] = Field(max_length=4096)


# Metadata base class
class DataRowMetadataField(_CamelCaseMixin):
    # One of `schema_id` or `name` must be provided. If `schema_id` is not provided, it is
    # inferred from `name`
    # schema id alias to json key name for pydantic v2 support
    schema_id: Optional[SchemaId] = None
    name: Optional[str] = None
    # value is of type `Any` so that we do not improperly coerce the value to the wrong type
    # Additional validation is performed before upload using the schema information
    value: Any


class DataRowMetadata(_CamelCaseMixin):
    global_key: Optional[str] = None
    data_row_id: Optional[str] = None
    fields: List[DataRowMetadataField]


class DeleteDataRowMetadata(_CamelCaseMixin):
    data_row_id: Union[str, UniqueId, GlobalKey] = None
    fields: List[SchemaId]


class DataRowMetadataBatchResponse(_CamelCaseMixin):
    global_key: Optional[str] = None
    data_row_id: Optional[str] = None
    error: Optional[str] = None
    fields: List[Union[DataRowMetadataField, SchemaId]]


# --- Batch GraphQL Objects ---
# Don't want to crowd the name space with internals


# Bulk upsert values
class _UpsertDataRowMetadataInput(_CamelCaseMixin):
    schema_id: str
    value: Any


# Batch of upsert values for a datarow
class _UpsertBatchDataRowMetadata(_CamelCaseMixin):
    global_key: Optional[str] = None
    data_row_id: Optional[str] = None
    fields: List[_UpsertDataRowMetadataInput]


class _DeleteBatchDataRowMetadata(_CamelCaseMixin):
    data_row_identifier: Union[UniqueId, GlobalKey]
    schema_ids: List[SchemaId]

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @model_serializer(mode="wrap")
    def model_serializer(self, handler):
        res = handler(self)
        if "data_row_identifier" in res.keys():
            key = "data_row_identifier"
            id_type_key = "id_type"
        else:
            key = "dataRowIdentifier"
            id_type_key = "idType"
        data_row_identifier = res.pop(key)
        res[key] = {
            "id": data_row_identifier.key,
            id_type_key: data_row_identifier.id_type,
        }
        return res


_BatchInputs = Union[
    List[_UpsertBatchDataRowMetadata], List[_DeleteBatchDataRowMetadata]
]
_BatchFunction = Callable[[_BatchInputs], List[DataRowMetadataBatchResponse]]


class _UpsertCustomMetadataSchemaEnumOptionInput(_CamelCaseMixin):
    id: Optional[SchemaId] = None
    name: Annotated[
        str,
        StringConstraints(strip_whitespace=True, min_length=1, max_length=100),
    ]
    kind: str


class _UpsertCustomMetadataSchemaInput(_CamelCaseMixin):
    id: Optional[SchemaId] = None
    name: Annotated[
        str,
        StringConstraints(strip_whitespace=True, min_length=1, max_length=100),
    ]
    kind: str
    options: Optional[List[_UpsertCustomMetadataSchemaEnumOptionInput]] = None


class DataRowMetadataOntology:
    """Ontology for data row metadata

    Metadata provides additional context for a data rows. Metadata is broken into two classes
    reserved and custom. Reserved fields are defined by Labelbox and used for creating
    specific experiences in the platform.

    >>> mdo = client.get_data_row_metadata_ontology()

    """

    def __init__(self, client):
        self._client = client
        self._batch_size = 50  # used for uploads and deletes

        self._raw_ontology = self._get_ontology()
        self._build_ontology()

    def _build_ontology(self):
        # all fields
        self.fields = self._parse_ontology(self._raw_ontology)
        self.fields_by_id = self._make_id_index(self.fields)

        # reserved fields
        self.reserved_fields: List[DataRowMetadataSchema] = [
            f for f in self.fields if f.reserved
        ]
        self.reserved_by_id = self._make_id_index(self.reserved_fields)
        self.reserved_by_name: Dict[
            str, Union[DataRowMetadataSchema, Dict[str, DataRowMetadataSchema]]
        ] = self._make_name_index(self.reserved_fields)
        self.reserved_by_name_normalized: Dict[str, DataRowMetadataSchema] = (
            self._make_normalized_name_index(self.reserved_fields)
        )

        # custom fields
        self.custom_fields: List[DataRowMetadataSchema] = [
            f for f in self.fields if not f.reserved
        ]
        self.custom_by_id = self._make_id_index(self.custom_fields)
        self.custom_by_name: Dict[
            str, Union[DataRowMetadataSchema, Dict[str, DataRowMetadataSchema]]
        ] = self._make_name_index(self.custom_fields)
        self.custom_by_name_normalized: Dict[str, DataRowMetadataSchema] = (
            self._make_normalized_name_index(self.custom_fields)
        )

    @staticmethod
    def _lookup_in_index_by_name(reserved_index, custom_index, name):
        # search through reserved names first
        if name in reserved_index:
            return reserved_index[name]
        elif name in custom_index:
            return custom_index[name]
        else:
            raise KeyError(f"There is no metadata with name '{name}'")

    def get_by_name(
        self, name: str
    ) -> Union[DataRowMetadataSchema, Dict[str, DataRowMetadataSchema]]:
        """Get metadata by name

        >>> mdo.get_by_name(name)

        Args:
            name (str): Name of metadata schema

        Returns:
            Metadata schema as `DataRowMetadataSchema` or dict, in case of Enum metadata

        Raises:
            KeyError: When provided name is not presented in neither reserved nor custom metadata list
        """
        return self._lookup_in_index_by_name(
            self.reserved_by_name, self.custom_by_name, name
        )

    def _get_by_name_normalized(self, name: str) -> DataRowMetadataSchema:
        """Get metadata by name. For options, it provides the option schema instead of list of
        options
        """
        # using `normalized` indices to find options by name as well
        return self._lookup_in_index_by_name(
            self.reserved_by_name_normalized,
            self.custom_by_name_normalized,
            name,
        )

    @staticmethod
    def _make_name_index(
        fields: List[DataRowMetadataSchema],
    ) -> Dict[
        str, Union[DataRowMetadataSchema, Dict[str, DataRowMetadataSchema]]
    ]:
        index = {}
        for f in fields:
            if f.options:
                index[f.name] = {}
                for o in f.options:
                    index[f.name][o.name] = o
            else:
                index[f.name] = f
        return index

    @staticmethod
    def _make_normalized_name_index(
        fields: List[DataRowMetadataSchema],
    ) -> Dict[str, DataRowMetadataSchema]:
        index = {}
        for f in fields:
            index[f.name] = f
        return index

    @staticmethod
    def _make_id_index(
        fields: List[DataRowMetadataSchema],
    ) -> Dict[SchemaId, DataRowMetadataSchema]:
        index = {}
        for f in fields:
            index[f.uid] = f
            if f.options:
                for o in f.options:
                    index[o.uid] = o
        return index

    def _get_ontology(self) -> List[Dict[str, Any]]:
        query = """query GetMetadataOntologyBetaPyApi {
        customMetadataOntology {
                id
                name
                kind
                reserved
                options {
                    id
                    kind
                    name
                    reserved
                }
        }}
        """
        return self._client.execute(query)["customMetadataOntology"]

    @staticmethod
    def _parse_ontology(raw_ontology) -> List[DataRowMetadataSchema]:
        fields = []
        copy = deepcopy(raw_ontology)
        for schema in copy:
            schema["uid"] = schema["id"]
            options = None
            if schema.get("options"):
                options = []
                for option in schema["options"]:
                    option["uid"] = option["id"]
                    options.append(
                        DataRowMetadataSchema(
                            **{**option, **{"parent": schema["uid"]}}
                        )
                    )
            schema["options"] = options
            fields.append(DataRowMetadataSchema(**schema))

        return fields

    def refresh_ontology(self):
        """Update the `DataRowMetadataOntology` instance with the latest
        metadata ontology schemas
        """
        self._raw_ontology = self._get_ontology()
        self._build_ontology()

    def create_schema(
        self, name: str, kind: DataRowMetadataKind, options: List[str] = None
    ) -> DataRowMetadataSchema:
        """Create metadata schema

        >>> mdo.create_schema(name, kind, options)

        Args:
            name (str): Name of metadata schema
            kind (DataRowMetadataKind): Kind of metadata schema as `DataRowMetadataKind`
            options (List[str]): List of Enum options

        Returns:
            Created metadata schema as `DataRowMetadataSchema`

        Raises:
            KeyError: When provided name is not a valid custom metadata
        """
        if not isinstance(kind, DataRowMetadataKind):
            raise ValueError(f"kind '{kind}' must be a `DataRowMetadataKind`")

        upsert_schema = _UpsertCustomMetadataSchemaInput(
            name=name, kind=kind.value
        )
        if options:
            if kind != DataRowMetadataKind.enum:
                raise ValueError(
                    f"Kind '{kind}' must be an Enum, if Enum options are provided"
                )
            upsert_enum_options = [
                _UpsertCustomMetadataSchemaEnumOptionInput(
                    name=o, kind=DataRowMetadataKind.option.value
                )
                for o in options
            ]
            upsert_schema.options = upsert_enum_options

        return self._upsert_schema(upsert_schema)

    def update_schema(self, name: str, new_name: str) -> DataRowMetadataSchema:
        """Update metadata schema

        >>> mdo.update_schema(name, new_name)

        Args:
            name (str): Current name of metadata schema
            new_name (str): New name of metadata schema

        Returns:
            Updated metadata schema as `DataRowMetadataSchema`

        Raises:
            KeyError: When provided name is not a valid custom metadata
        """
        schema = self._validate_custom_schema_by_name(name)
        upsert_schema = _UpsertCustomMetadataSchemaInput(
            id=schema.uid, name=new_name, kind=schema.kind.value
        )
        if schema.options:
            upsert_enum_options = [
                _UpsertCustomMetadataSchemaEnumOptionInput(
                    id=o.uid, name=o.name, kind=DataRowMetadataKind.option.value
                )
                for o in schema.options
            ]
            upsert_schema.options = upsert_enum_options

        return self._upsert_schema(upsert_schema)

    def update_enum_option(
        self, name: str, option: str, new_option: str
    ) -> DataRowMetadataSchema:
        """Update Enum metadata schema option

        >>> mdo.update_enum_option(name, option, new_option)

        Args:
            name (str): Name of metadata schema to update
            option (str): Name of Enum option to update
            new_option (str): New name of Enum option

        Returns:
            Updated metadata schema as `DataRowMetadataSchema`

        Raises:
            KeyError: When provided name is not a valid custom metadata
        """
        schema = self._validate_custom_schema_by_name(name)
        if schema.kind != DataRowMetadataKind.enum:
            raise ValueError(
                "Updating Enum option is only supported for Enum metadata schema"
            )
        valid_options: List[str] = [o.name for o in schema.options]

        if option not in valid_options:
            raise ValueError(
                f"Enum option '{option}' is not a valid option for Enum '{name}', valid options are: {valid_options}"
            )
        upsert_schema = _UpsertCustomMetadataSchemaInput(
            id=schema.uid, name=schema.name, kind=schema.kind.value
        )
        upsert_enum_options = []
        for o in schema.options:
            enum_option = _UpsertCustomMetadataSchemaEnumOptionInput(
                id=o.uid, name=o.name, kind=o.kind.value
            )
            if enum_option.name == option:
                enum_option.name = new_option
            upsert_enum_options.append(enum_option)
        upsert_schema.options = upsert_enum_options

        return self._upsert_schema(upsert_schema)

    def delete_schema(self, name: str) -> bool:
        """Delete metadata schema

        >>> mdo.delete_schema(name)

        Args:
            name: Name of metadata schema to delete

        Returns:
            True if deletion is successful, False if unsuccessful

        Raises:
            KeyError: When provided name is not a valid custom metadata
        """
        schema = self._validate_custom_schema_by_name(name)
        query = """mutation DeleteCustomMetadataSchemaPyApi($where: WhereUniqueIdInput!) {
                deleteCustomMetadataSchema(schema: $where){
                    success
                }
            }"""
        res = self._client.execute(query, {"where": {"id": schema.uid}})[
            "deleteCustomMetadataSchema"
        ]
        self.refresh_ontology()

        return res["success"]

    def parse_metadata(
        self, unparsed: List[Dict[str, List[Union[str, Dict]]]]
    ) -> List[DataRowMetadata]:
        """Parse metadata responses

        >>> mdo.parse_metadata([metadata])

        Args:
            unparsed: An unparsed metadata export

        Returns:
            metadata: List of `DataRowMetadata`
        """
        parsed = []
        if isinstance(unparsed, dict):
            raise ValueError("Pass a list of dictionaries")

        for dr in unparsed:
            fields = []
            if "fields" in dr:
                fields = self.parse_metadata_fields(dr["fields"])
            parsed.append(
                DataRowMetadata(
                    data_row_id=dr["dataRowId"],
                    global_key=dr["globalKey"],
                    fields=fields,
                )
            )
        return parsed

    def parse_metadata_fields(
        self, unparsed: List[Dict[str, Dict]]
    ) -> List[DataRowMetadataField]:
        """Parse metadata fields as list of `DataRowMetadataField`

        >>> mdo.parse_metadata_fields([metadata_fields])

        Args:
            unparsed: An unparsed list of metadata represented as a dict containing 'schemaId' and 'value'

        Returns:
            metadata: List of `DataRowMetadataField`
        """
        parsed = []
        if isinstance(unparsed, dict):
            raise ValueError("Pass a list of dictionaries")

        for f in unparsed:
            if f["schemaId"] not in self.fields_by_id:
                # Fetch latest metadata ontology if metadata can't be found
                self.refresh_ontology()
                if f["schemaId"] not in self.fields_by_id:
                    raise ValueError(
                        f"Schema Id `{f['schemaId']}` not found in ontology"
                    )

            schema = self.fields_by_id[f["schemaId"]]
            if schema.kind == DataRowMetadataKind.enum:
                continue
            elif schema.kind == DataRowMetadataKind.option:
                field = DataRowMetadataField(
                    schema_id=schema.parent, value=schema.uid
                )
            elif schema.kind == DataRowMetadataKind.datetime:
                field = DataRowMetadataField(
                    schema_id=schema.uid,
                    value=format_iso_from_string(f["value"]),
                )
            else:
                field = DataRowMetadataField(
                    schema_id=schema.uid, value=f["value"]
                )

            field.name = schema.name
            parsed.append(field)
        return parsed

    def bulk_upsert(
        self, metadata: List[DataRowMetadata]
    ) -> List[DataRowMetadataBatchResponse]:
        """Upsert metadata to a list of data rows

        You may specify data row by either data_row_id or global_key

        >>> metadata = DataRowMetadata(
        >>>                 data_row_id="datarow-id", # Alternatively, set global_key="global-key"
        >>>                 fields=[
        >>>                        DataRowMetadataField(schema_id="schema-id", value="my-message"),
        >>>                        ...
        >>>                    ]
        >>>    )
        >>> mdo.batch_upsert([metadata])

        Args:
            metadata: List of DataRow Metadata to upsert

        Returns:
            list of unsuccessful upserts.
            An empty list means the upload was successful.
        """

        if not len(metadata):
            raise ValueError("Empty list passed")

        def _batch_upsert(
            upserts: List[_UpsertBatchDataRowMetadata],
        ) -> List[DataRowMetadataBatchResponse]:
            query = """mutation UpsertDataRowMetadataBetaPyApi($metadata: [DataRowCustomMetadataBatchUpsertInput!]!) {
                upsertDataRowCustomMetadata(data: $metadata){
                    globalKey
                    dataRowId
                    error
                    fields {
                        value
                        schemaId
                    }
                }
            }"""
            res = self._client.execute(query, {"metadata": upserts})[
                "upsertDataRowCustomMetadata"
            ]
            return [
                DataRowMetadataBatchResponse(
                    global_key=r["globalKey"],
                    data_row_id=r["dataRowId"],
                    error=r["error"],
                    fields=self.parse_metadata([r])[0].fields,
                )
                for r in res
            ]

        items = []
        for m in metadata:
            items.append(
                _UpsertBatchDataRowMetadata(
                    global_key=m.global_key,
                    data_row_id=m.data_row_id,
                    fields=list(
                        chain.from_iterable(
                            self._parse_upsert(f, m.data_row_id)
                            for f in m.fields
                        )
                    ),
                ).model_dump(by_alias=True)
            )
        res = _batch_operations(_batch_upsert, items, self._batch_size)
        return res

    def bulk_delete(
        self, deletes: List[DeleteDataRowMetadata]
    ) -> List[DataRowMetadataBatchResponse]:
        """Delete metadata from a datarow by specifiying the fields you want to remove

        >>> delete = DeleteDataRowMetadata(
        >>>                 data_row_id=UniqueId("datarow-id"),
        >>>                 fields=[
        >>>                        "schema-id-1",
        >>>                        "schema-id-2"
        >>>                        ...
        >>>                    ]
        >>>    )
        >>> mdo.batch_delete([metadata])

        >>> delete = DeleteDataRowMetadata(
        >>>                 data_row_id=GlobalKey("global-key"),
        >>>                 fields=[
        >>>                        "schema-id-1",
        >>>                        "schema-id-2"
        >>>                        ...
        >>>                    ]
        >>>    )
        >>> mdo.batch_delete([metadata])

        >>> delete = DeleteDataRowMetadata(
        >>>                 data_row_id="global-key",
        >>>                 fields=[
        >>>                        "schema-id-1",
        >>>                        "schema-id-2"
        >>>                        ...
        >>>                    ]
        >>>    )
        >>> mdo.batch_delete([metadata])


        Args:
            deletes: Data row and schema ids to delete
                For data row, we support UniqueId, str, and GlobalKey.
                If you pass a str, we will assume it is a UniqueId
                Do not pass a mix of data row ids and global keys in the same list

        Returns:
            list of unsuccessful deletions.
            An empty list means all data rows were successfully deleted.

        """

        if not len(deletes):
            raise ValueError("The 'deletes' list cannot be empty.")

        for i, delete in enumerate(deletes):
            if isinstance(delete.data_row_id, str):
                deletes[i] = DeleteDataRowMetadata(
                    data_row_id=UniqueId(delete.data_row_id),
                    fields=delete.fields,
                )
            elif isinstance(delete.data_row_id, UniqueId):
                continue
            elif isinstance(delete.data_row_id, GlobalKey):
                continue
            else:
                raise ValueError(
                    f"Invalid data row identifier type '{type(delete.data_row_id)}' for '{delete.data_row_id}'"
                )

        def _batch_delete(
            deletes: List[_DeleteBatchDataRowMetadata],
        ) -> List[DataRowMetadataBatchResponse]:
            query = """mutation DeleteDataRowMetadataBetaPyApi($deletes: [DataRowIdentifierCustomMetadataBatchDeleteInput!]) {
                deleteDataRowCustomMetadata(dataRowIdentifiers: $deletes) {
                    dataRowId
                    error
                    fields {
                        value
                        schemaId
                    }
                }
            }
            """
            res = self._client.execute(query, {"deletes": deletes})[
                "deleteDataRowCustomMetadata"
            ]
            failures = []
            for dr in res:
                dr["fields"] = [f["schemaId"] for f in dr["fields"]]
                failures.append(DataRowMetadataBatchResponse(**dr))
            return failures

        items = [self._validate_delete(m) for m in deletes]
        return _batch_operations(
            _batch_delete, items, batch_size=self._batch_size
        )

    @overload
    def bulk_export(self, data_row_ids: List[str]) -> List[DataRowMetadata]:
        pass

    @overload
    def bulk_export(
        self, data_row_ids: DataRowIdentifiers
    ) -> List[DataRowMetadata]:
        pass

    def bulk_export(self, data_row_ids) -> List[DataRowMetadata]:
        """Exports metadata for a list of data rows

        >>> mdo.bulk_export([data_row.uid for data_row in data_rows])

        Args:
            data_row_ids: List of data data rows to fetch metadata for. This can be a list of strings or a DataRowIdentifiers object
            DataRowIdentifier objects are lists of ids or global keys. A DataIdentifier object can be a UniqueIds or GlobalKeys class.
        Returns:
            A list of DataRowMetadata.
            There will be one DataRowMetadata for each data_row_id passed in.
            This is true even if the data row does not have any meta data.
            Data rows without metadata will have empty `fields`.

        """
        if not len(data_row_ids):
            raise ValueError("Empty list passed")

        if (
            isinstance(data_row_ids, list)
            and len(data_row_ids) > 0
            and isinstance(data_row_ids[0], str)
        ):
            data_row_ids = UniqueIds(data_row_ids)

        def _bulk_export(
            _data_row_ids: DataRowIdentifiers,
        ) -> List[DataRowMetadata]:
            query = """query dataRowCustomMetadataPyApi($dataRowIdentifiers: DataRowCustomMetadataDataRowIdentifiersInput) {
                dataRowCustomMetadata(where: {dataRowIdentifiers : $dataRowIdentifiers}) {
                    dataRowId
                    globalKey
                    fields {
                        value
                        schemaId
                    }
                }
            }
            """
            return self.parse_metadata(
                self._client.execute(
                    query,
                    {
                        "dataRowIdentifiers": {
                            "ids": [id for id in _data_row_ids],
                            "idType": _data_row_ids.id_type,
                        }
                    },
                )["dataRowCustomMetadata"]
            )

        return _batch_operations(
            _bulk_export, data_row_ids, batch_size=self._batch_size
        )

    def parse_upsert_metadata(self, metadata_fields) -> List[Dict[str, Any]]:
        """Converts either `DataRowMetadataField` or a dictionary representation
            of `DataRowMetadataField` into a validated, flattened dictionary of
            metadata fields that are used to create data row metadata. Used
            internally in `Dataset.create_data_rows()`

        Args:
            metadata_fields: List of `DataRowMetadataField` or a dictionary representation
                of `DataRowMetadataField`
        Returns:
            List of dictionaries representing a flattened view of metadata fields
        """

        def _convert_metadata_field(metadata_field):
            if isinstance(metadata_field, DataRowMetadataField):
                return metadata_field
            elif isinstance(metadata_field, dict):
                if "value" not in metadata_field:
                    raise ValueError(
                        f"Custom metadata field '{metadata_field}' must have a 'value' key"
                    )
                if (
                    "schema_id" not in metadata_field
                    and "name" not in metadata_field
                ):
                    raise ValueError(
                        f"Custom metadata field '{metadata_field}' must have either 'schema_id' or 'name' key"
                    )
                return DataRowMetadataField(
                    schema_id=metadata_field.get("schema_id"),
                    name=metadata_field.get("name"),
                    value=metadata_field["value"],
                )
            else:
                raise ValueError(
                    f"Metadata field '{metadata_field}' is neither 'DataRowMetadataField' type or a dictionary"
                )

        # Convert all metadata fields to DataRowMetadataField type
        metadata_fields = [_convert_metadata_field(m) for m in metadata_fields]
        parsed_metadata = list(
            chain.from_iterable(self._parse_upsert(m) for m in metadata_fields)
        )
        return [m.model_dump(by_alias=True) for m in parsed_metadata]

    def _upsert_schema(
        self, upsert_schema: _UpsertCustomMetadataSchemaInput
    ) -> DataRowMetadataSchema:
        query = """mutation UpsertCustomMetadataSchemaPyApi($data: UpsertCustomMetadataSchemaInput!) {
                upsertCustomMetadataSchema(data: $data){
                    id
                    name
                    kind
                    options {
                        id
                        name
                        kind
                    }
                }
            }"""
        res = self._client.execute(
            query, {"data": upsert_schema.model_dump(exclude_none=True)}
        )["upsertCustomMetadataSchema"]
        self.refresh_ontology()
        return _parse_metadata_schema(res)

    def _load_option_by_name(self, metadatum: DataRowMetadataField):
        is_value_a_valid_schema_id = metadatum.value in self.fields_by_id
        if not is_value_a_valid_schema_id:
            metadatum_by_name = self.get_by_name(metadatum.name)
            if metadatum.value not in metadatum_by_name:
                raise KeyError(
                    f"There is no enum option by name '{metadatum.value}' for enum name '{metadatum.name}'"
                )
            metadatum.value = metadatum_by_name[metadatum.value].uid

    def _load_schema_id_by_name(self, metadatum: DataRowMetadataField):
        """
        Loads schema id by name for a metadata field including options schema id.
        """
        if metadatum.name is None:
            return

        if metadatum.schema_id is None:
            schema = self._get_by_name_normalized(metadatum.name)
            metadatum.schema_id = schema.uid
            if schema.options:
                self._load_option_by_name(metadatum)

    def _parse_upsert(
        self, metadatum: DataRowMetadataField, data_row_id: Optional[str] = None
    ) -> List[_UpsertDataRowMetadataInput]:
        """Format for metadata upserts to GQL"""

        self._load_schema_id_by_name(metadatum)

        if metadatum.schema_id not in self.fields_by_id:
            # Fetch latest metadata ontology if metadata can't be found
            self.refresh_ontology()
            if metadatum.schema_id not in self.fields_by_id:
                raise ValueError(
                    f"Schema Id `{metadatum.schema_id}` not found in ontology"
                )

        schema = self.fields_by_id[metadatum.schema_id]
        try:
            if schema.kind == DataRowMetadataKind.datetime:
                parsed = _validate_parse_datetime(metadatum)
            elif schema.kind == DataRowMetadataKind.string:
                parsed = _validate_parse_text(metadatum)
            elif schema.kind == DataRowMetadataKind.number:
                parsed = _validate_parse_number(metadatum)
            elif schema.kind == DataRowMetadataKind.embedding:
                parsed = _validate_parse_embedding(metadatum)
            elif schema.kind == DataRowMetadataKind.enum:
                parsed = _validate_enum_parse(schema, metadatum)
            elif schema.kind == DataRowMetadataKind.option:
                raise ValueError(
                    "An Option id should not be set as the Schema id"
                )
            else:
                raise ValueError(f"Unknown type: {schema}")
        except ValueError as e:
            error_str = f"Could not validate metadata [{metadatum}]"
            if data_row_id:
                error_str += f", data_row_id='{data_row_id}'"
            raise ValueError(f"{error_str}. Reason: {e}")
        return [_UpsertDataRowMetadataInput(**p) for p in parsed]

    def _validate_delete(self, delete: DeleteDataRowMetadata):
        if not len(delete.fields):
            raise ValueError(f"No fields specified for {delete.data_row_id}")

        deletes = set()
        for schema_id in delete.fields:
            if schema_id not in self.fields_by_id:
                # Fetch latest metadata ontology if metadata can't be found
                self.refresh_ontology()
                if schema_id not in self.fields_by_id:
                    raise ValueError(
                        f"Schema Id `{schema_id}` not found in ontology"
                    )

            schema = self.fields_by_id[schema_id]
            # handle users specifying enums by adding all option enums
            if schema.kind == DataRowMetadataKind.enum:
                [deletes.add(o.uid) for o in schema.options]

            deletes.add(schema.uid)

        return _DeleteBatchDataRowMetadata(
            data_row_identifier=delete.data_row_id,
            schema_ids=list(delete.fields),
        ).model_dump(by_alias=True)

    def _validate_custom_schema_by_name(
        self, name: str
    ) -> DataRowMetadataSchema:
        if name not in self.custom_by_name_normalized:
            # Fetch latest metadata ontology if metadata can't be found
            self.refresh_ontology()
            if name not in self.custom_by_name_normalized:
                raise KeyError(f"'{name}' is not a valid custom metadata")

        return self.custom_by_name_normalized[name]


def _batch_items(iterable: List[Any], size: int) -> Generator[Any, None, None]:
    for ndx in range(0, len(iterable), size):
        yield iterable[ndx : min(ndx + size, len(iterable))]


def _batch_operations(
    batch_function: _BatchFunction,
    items: List,
    batch_size: int = 100,
):
    response = []

    for batch in _batch_items(items, batch_size):
        response += batch_function(batch)
    return response


def _validate_parse_embedding(
    field: DataRowMetadataField,
) -> List[Dict[str, Union[SchemaId, Embedding]]]:
    if isinstance(field.value, list):
        if not (Embedding.min_items <= len(field.value) <= Embedding.max_items):
            raise ValueError(
                "Embedding length invalid. "
                "Must have length within the interval "
                f"[{Embedding.min_items},{Embedding.max_items}]. Found {len(field.value)}"
            )
        field.value = [float(x) for x in field.value]
    else:
        raise ValueError(
            f"Expected a list for embedding. Found {type(field.value)}"
        )
    return [field.model_dump(by_alias=True)]


def _validate_parse_number(
    field: DataRowMetadataField,
) -> List[Dict[str, Union[SchemaId, str, float, int]]]:
    field.value = float(field.value)
    return [field.model_dump(by_alias=True)]


def _validate_parse_datetime(
    field: DataRowMetadataField,
) -> List[Dict[str, Union[SchemaId, str]]]:
    if isinstance(field.value, str):
        field.value = format_iso_from_string(field.value)
    elif not isinstance(field.value, datetime):
        raise TypeError(
            f"Value for datetime fields must be either a string or datetime object. Found {type(field.value)}"
        )

    return [
        {"schemaId": field.schema_id, "value": format_iso_datetime(field.value)}
    ]


def _validate_parse_text(
    field: DataRowMetadataField,
) -> List[Dict[str, Union[SchemaId, str]]]:
    if not isinstance(field.value, str):
        raise ValueError(
            f"Expected a string type for the text field. Found {type(field.value)}"
        )
    if len(field.value) > String.metadata[0].max_length:
        raise ValueError(
            f"String fields cannot exceed {String.metadata.max_length} characters."
        )
    return [field.model_dump(by_alias=True)]


def _validate_enum_parse(
    schema: DataRowMetadataSchema, field: DataRowMetadataField
) -> List[Dict[str, Union[SchemaId, dict]]]:
    if schema.options:
        if field.value not in {o.uid for o in schema.options}:
            raise ValueError(
                f"Option `{field.value}` not found for {field.schema_id}"
            )
    else:
        raise ValueError("Incorrectly specified enum schema")

    return [
        {"schemaId": field.schema_id, "value": {}},
        {"schemaId": field.value, "value": {}},
    ]


def _parse_metadata_schema(
    unparsed: Dict[str, Union[str, List]],
) -> DataRowMetadataSchema:
    uid = unparsed["id"]
    name = unparsed["name"]
    kind = DataRowMetadataKind(unparsed["kind"])
    options = [
        DataRowMetadataSchema(
            uid=o["id"],
            name=o["name"],
            reserved=False,
            kind=DataRowMetadataKind.option,
            parent=uid,
        )
        for o in unparsed["options"]
    ]
    return DataRowMetadataSchema(
        uid=uid, name=name, reserved=False, kind=kind, options=options or None
    )



import time
import os
import pytest

from collections import Counter

from labelbox import DataSplit, ModelRun


@pytest.fixture
def current_model(client, configured_project_with_label, rand_gen):
    project, _, _, label = configured_project_with_label
    ontology = project.ontology()

    model = client.create_model(rand_gen(str), ontology.uid)
    yield model

    model.delete()


def test_model_run(
    client, configured_project_with_label, current_model, data_row, rand_gen
):
    _, _, _, label = configured_project_with_label
    label_id = label.uid
    model = current_model

    name = rand_gen(str)
    config = {"batch_size": 100, "reruns": None}
    model_run = model.create_model_run(name, config)
    assert model_run.name == name
    assert model_run.training_metadata["batchSize"] == config["batch_size"]
    assert model_run.training_metadata["reruns"] == config["reruns"]
    assert model_run.model_id == model.uid
    assert model_run.created_by_id == client.get_user().uid

    model_run.upsert_labels([label_id])
    time.sleep(3)

    model_run_data_row = next(model_run.model_run_data_rows())
    assert model_run_data_row.label_id == label_id
    assert model_run_data_row.model_run_id == model_run.uid
    assert model_run_data_row.data_row().uid == data_row.uid

    fetch_model_run = client.get_model_run(model_run.uid)
    assert fetch_model_run == model_run


def test_model_run_no_config(rand_gen, model):
    name = rand_gen(str)
    model_run = model.create_model_run(name)
    assert model_run.name == name


def test_model_run_delete(client, model_run):
    models_before = list(client.get_models())
    model_before = models_before[0]
    before = list(model_before.model_runs())

    model_run = before[0]
    model_run.delete()

    models_after = list(client.get_models())
    model_after = models_after[0]
    after = list(model_after.model_runs())
    after_uids = {mr.uid for mr in after}

    assert model_run.uid not in after_uids


def test_model_run_update_config(model_run_with_training_metadata):
    new_config = {"batch_size": 2000}
    res = model_run_with_training_metadata.update_config(new_config)
    assert res["trainingMetadata"]["batch_size"] == new_config["batch_size"]


def test_model_run_reset_config(model_run_with_training_metadata):
    res = model_run_with_training_metadata.reset_config()
    assert res["trainingMetadata"] is None


def test_model_run_get_config(model_run_with_training_metadata):
    new_config = {"batch_size": 2000}
    model_run_with_training_metadata.update_config(new_config)
    res = model_run_with_training_metadata.get_config()
    assert res["batch_size"] == new_config["batch_size"]


def test_model_run_data_rows_delete(model_run_with_data_rows):
    model_run = model_run_with_data_rows

    before = list(model_run.model_run_data_rows())
    annotation_data_row = before[0]

    data_row_id = annotation_data_row.data_row().uid
    model_run.delete_model_run_data_rows(data_row_ids=[data_row_id])
    after = list(model_run.model_run_data_rows())
    assert len(before) == len(after) + 1


def test_model_run_upsert_data_rows(dataset, model_run, configured_project):
    n_model_run_data_rows = len(list(model_run.model_run_data_rows()))
    assert n_model_run_data_rows == 0
    data_row = dataset.create_data_row(row_data="test row data")
    configured_project._wait_until_data_rows_are_processed(
        data_row_ids=[data_row.uid]
    )
    model_run.upsert_data_rows([data_row.uid])
    n_model_run_data_rows = len(list(model_run.model_run_data_rows()))
    assert n_model_run_data_rows == 1


@pytest.mark.parametrize("data_rows", [2], indirect=True)
def test_model_run_upsert_data_rows_using_global_keys(model_run, data_rows):
    global_keys = [dr.global_key for dr in data_rows]
    assert model_run.upsert_data_rows(global_keys=global_keys)
    model_run_data_rows = list(model_run.model_run_data_rows())
    added_data_rows = [mdr.data_row() for mdr in model_run_data_rows]
    assert set(added_data_rows) == set(data_rows)


def test_model_run_upsert_data_rows_with_existing_labels(
    model_run_with_data_rows,
):
    model_run_data_rows = list(model_run_with_data_rows.model_run_data_rows())
    n_data_rows = len(model_run_data_rows)
    model_run_with_data_rows.upsert_data_rows(
        [
            model_run_data_row.data_row().uid
            for model_run_data_row in model_run_data_rows
        ]
    )
    assert n_data_rows == len(
        list(model_run_with_data_rows.model_run_data_rows())
    )


@pytest.mark.skipif(
    condition=os.environ["LABELBOX_TEST_ENVIRON"] == "onprem",
    reason="does not work for onprem",
)
def test_model_run_status(model_run_with_data_rows):
    def get_model_run_status():
        return model_run_with_data_rows.client.execute(
            """query trainingPipelinePyApi($modelRunId: ID!) {
            trainingPipeline(where: {id : $modelRunId}) {status, errorMessage, metadata}}
        """,
            {"modelRunId": model_run_with_data_rows.uid},
            experimental=True,
        )["trainingPipeline"]

    model_run_status = get_model_run_status()
    assert model_run_status["status"] is None
    assert model_run_status["metadata"] is None
    assert model_run_status["errorMessage"] is None

    status = "COMPLETE"
    metadata = {"key1": "value1"}
    errorMessage = "an error"
    model_run_with_data_rows.update_status(status, metadata, errorMessage)

    model_run_status = get_model_run_status()
    assert model_run_status["status"] == status
    assert model_run_status["metadata"] == metadata
    assert model_run_status["errorMessage"] == errorMessage

    extra_metadata = {"key2": "value2"}
    model_run_with_data_rows.update_status(status, extra_metadata)
    model_run_status = get_model_run_status()
    assert model_run_status["status"] == status
    assert model_run_status["metadata"] == {**metadata, **extra_metadata}
    assert model_run_status["errorMessage"] == errorMessage

    status = ModelRun.Status.FAILED
    model_run_with_data_rows.update_status(status, metadata, errorMessage)
    model_run_status = get_model_run_status()
    assert model_run_status["status"] == status.value

    with pytest.raises(ValueError):
        model_run_with_data_rows.update_status(
            "INVALID", metadata, errorMessage
        )


def test_model_run_split_assignment_by_data_row_ids(
    model_run, dataset, image_url
):
    n_data_rows = 2
    data_rows = dataset.create_data_rows(
        [{"row_data": image_url} for _ in range(n_data_rows)]
    )
    data_rows.wait_till_done()
    data_row_ids = [data_row["id"] for data_row in data_rows.result]
    model_run.upsert_data_rows(data_row_ids)

    with pytest.raises(ValueError):
        model_run.assign_data_rows_to_split(data_row_ids, "INVALID SPLIT")

    for split in ["TRAINING", "TEST", "VALIDATION", "UNASSIGNED", *DataSplit]:
        model_run.assign_data_rows_to_split(data_row_ids, split)
        counts = Counter()
        for data_row in model_run.model_run_data_rows():
            counts[data_row.data_split.value] += 1
        split = split.value if isinstance(split, DataSplit) else split
        assert counts[split] == n_data_rows


@pytest.mark.parametrize("data_rows", [2], indirect=True)
def test_model_run_split_assignment_by_global_keys(model_run, data_rows):
    global_keys = [data_row.global_key for data_row in data_rows]

    model_run.upsert_data_rows(global_keys=global_keys)

    for split in ["TRAINING", "TEST", "VALIDATION", "UNASSIGNED", *DataSplit]:
        model_run.assign_data_rows_to_split(
            split=split, global_keys=global_keys
        )
        splits = [
            data_row.data_split.value
            for data_row in model_run.model_run_data_rows()
        ]
        assert len(set(splits)) == 1



import json
from labelbox.data.annotation_types.classification.classification import (
    Checklist,
    ClassificationAnnotation,
    ClassificationAnswer,
    Radio,
    Text,
)
from labelbox.data.annotation_types.data import GenericDataRowData
from labelbox.data.annotation_types.geometry.line import Line
from labelbox.data.annotation_types.geometry.point import Point
from labelbox.data.annotation_types.geometry.rectangle import Rectangle

from labelbox.data.annotation_types.label import Label
from labelbox.data.annotation_types.video import (
    VideoClassificationAnnotation,
    VideoObjectAnnotation,
)

from labelbox.data.serialization.ndjson.converter import NDJsonConverter
from operator import itemgetter


def test_video():
    with open("tests/data/assets/ndjson/video_import.json", "r") as file:
        data = json.load(file)

    labels = [
        Label(
            data=GenericDataRowData(uid="ckrb1sf1i1g7i0ybcdc6oc8ct"),
            annotations=[
                VideoClassificationAnnotation(
                    feature_schema_id="ckrb1sfjx099a0y914hl319ie",
                    extra={"uuid": "f6879f59-d2b5-49c2-aceb-d9e8dc478673"},
                    value=Radio(
                        answer=ClassificationAnswer(
                            feature_schema_id="ckrb1sfl8099g0y91cxbd5ftb",
                        ),
                    ),
                    frame=30,
                ),
                VideoClassificationAnnotation(
                    feature_schema_id="ckrb1sfjx099a0y914hl319ie",
                    extra={"uuid": "f6879f59-d2b5-49c2-aceb-d9e8dc478673"},
                    value=Radio(
                        answer=ClassificationAnswer(
                            feature_schema_id="ckrb1sfl8099g0y91cxbd5ftb",
                        ),
                    ),
                    frame=31,
                ),
                VideoClassificationAnnotation(
                    feature_schema_id="ckrb1sfjx099a0y914hl319ie",
                    extra={"uuid": "f6879f59-d2b5-49c2-aceb-d9e8dc478673"},
                    value=Radio(
                        answer=ClassificationAnswer(
                            feature_schema_id="ckrb1sfl8099g0y91cxbd5ftb",
                        ),
                    ),
                    frame=32,
                ),
                VideoClassificationAnnotation(
                    feature_schema_id="ckrb1sfjx099a0y914hl319ie",
                    extra={"uuid": "f6879f59-d2b5-49c2-aceb-d9e8dc478673"},
                    value=Radio(
                        answer=ClassificationAnswer(
                            feature_schema_id="ckrb1sfl8099g0y91cxbd5ftb",
                        ),
                    ),
                    frame=33,
                ),
                VideoClassificationAnnotation(
                    feature_schema_id="ckrb1sfjx099a0y914hl319ie",
                    extra={"uuid": "f6879f59-d2b5-49c2-aceb-d9e8dc478673"},
                    value=Radio(
                        answer=ClassificationAnswer(
                            feature_schema_id="ckrb1sfl8099g0y91cxbd5ftb",
                        ),
                    ),
                    frame=34,
                ),
                VideoClassificationAnnotation(
                    feature_schema_id="ckrb1sfjx099a0y914hl319ie",
                    extra={"uuid": "f6879f59-d2b5-49c2-aceb-d9e8dc478673"},
                    value=Radio(
                        answer=ClassificationAnswer(
                            feature_schema_id="ckrb1sfl8099g0y91cxbd5ftb",
                        ),
                    ),
                    frame=35,
                ),
                VideoClassificationAnnotation(
                    feature_schema_id="ckrb1sfjx099a0y914hl319ie",
                    extra={"uuid": "f6879f59-d2b5-49c2-aceb-d9e8dc478673"},
                    value=Radio(
                        answer=ClassificationAnswer(
                            feature_schema_id="ckrb1sfl8099g0y91cxbd5ftb",
                        ),
                    ),
                    frame=50,
                ),
                VideoClassificationAnnotation(
                    feature_schema_id="ckrb1sfjx099a0y914hl319ie",
                    extra={"uuid": "f6879f59-d2b5-49c2-aceb-d9e8dc478673"},
                    value=Radio(
                        answer=ClassificationAnswer(
                            feature_schema_id="ckrb1sfl8099g0y91cxbd5ftb",
                        ),
                    ),
                    frame=51,
                ),
                VideoClassificationAnnotation(
                    feature_schema_id="ckrb1sfkn099c0y910wbo0p1a",
                    extra={"uuid": "d009925d-91a3-4f67-abd9-753453f5a584"},
                    value=Checklist(
                        answer=[
                            ClassificationAnswer(
                                feature_schema_id="ckrb1sfl8099e0y919v260awv",
                            )
                        ],
                    ),
                    frame=0,
                ),
                VideoClassificationAnnotation(
                    feature_schema_id="ckrb1sfkn099c0y910wbo0p1a",
                    extra={"uuid": "d009925d-91a3-4f67-abd9-753453f5a584"},
                    value=Checklist(
                        answer=[
                            ClassificationAnswer(
                                feature_schema_id="ckrb1sfl8099e0y919v260awv",
                            )
                        ],
                    ),
                    frame=1,
                ),
                VideoClassificationAnnotation(
                    feature_schema_id="ckrb1sfkn099c0y910wbo0p1a",
                    extra={"uuid": "d009925d-91a3-4f67-abd9-753453f5a584"},
                    value=Checklist(
                        answer=[
                            ClassificationAnswer(
                                feature_schema_id="ckrb1sfl8099e0y919v260awv",
                            )
                        ],
                    ),
                    frame=2,
                ),
                VideoClassificationAnnotation(
                    feature_schema_id="ckrb1sfkn099c0y910wbo0p1a",
                    extra={"uuid": "d009925d-91a3-4f67-abd9-753453f5a584"},
                    value=Checklist(
                        answer=[
                            ClassificationAnswer(
                                feature_schema_id="ckrb1sfl8099e0y919v260awv",
                            )
                        ],
                    ),
                    frame=3,
                ),
                VideoClassificationAnnotation(
                    feature_schema_id="ckrb1sfkn099c0y910wbo0p1a",
                    extra={"uuid": "d009925d-91a3-4f67-abd9-753453f5a584"},
                    value=Checklist(
                        answer=[
                            ClassificationAnswer(
                                feature_schema_id="ckrb1sfl8099e0y919v260awv",
                            )
                        ],
                    ),
                    frame=4,
                ),
                VideoClassificationAnnotation(
                    feature_schema_id="ckrb1sfkn099c0y910wbo0p1a",
                    extra={"uuid": "d009925d-91a3-4f67-abd9-753453f5a584"},
                    value=Checklist(
                        answer=[
                            ClassificationAnswer(
                                feature_schema_id="ckrb1sfl8099e0y919v260awv",
                            )
                        ],
                    ),
                    frame=5,
                ),
                ClassificationAnnotation(
                    feature_schema_id="ckrb1sfkn099c0y910wbo0p1a",
                    extra={"uuid": "90e2ecf7-c19c-47e6-8cdb-8867e1b9d88c"},
                    value=Text(answer="a value"),
                ),
                VideoObjectAnnotation(
                    feature_schema_id="cl5islwg200gfci6g0oitaypu",
                    extra={"uuid": "6f7c835a-0139-4896-b73f-66a6baa89e94"},
                    value=Line(
                        points=[
                            Point(x=10.0, y=10.0),
                            Point(x=100.0, y=100.0),
                            Point(x=50.0, y=30.0),
                        ],
                    ),
                    frame=1,
                    keyframe=True,
                    segment_index=0,
                ),
                VideoObjectAnnotation(
                    feature_schema_id="cl5islwg200gfci6g0oitaypu",
                    extra={"uuid": "6f7c835a-0139-4896-b73f-66a6baa89e94"},
                    value=Line(
                        points=[
                            Point(x=15.0, y=10.0),
                            Point(x=50.0, y=100.0),
                            Point(x=50.0, y=30.0),
                        ],
                    ),
                    frame=5,
                    keyframe=True,
                    segment_index=0,
                ),
                VideoObjectAnnotation(
                    feature_schema_id="cl5islwg200gfci6g0oitaypu",
                    extra={"uuid": "6f7c835a-0139-4896-b73f-66a6baa89e94"},
                    value=Line(
                        points=[
                            Point(x=100.0, y=10.0),
                            Point(x=50.0, y=100.0),
                            Point(x=50.0, y=30.0),
                        ],
                    ),
                    frame=8,
                    keyframe=True,
                    segment_index=1,
                ),
                VideoObjectAnnotation(
                    feature_schema_id="cl5it7ktp00i5ci6gf80b1ysd",
                    extra={"uuid": "f963be22-227b-4efe-9be4-2738ed822216"},
                    value=Point(x=10.0, y=10.0),
                    frame=1,
                    keyframe=True,
                    segment_index=0,
                ),
                VideoObjectAnnotation(
                    feature_schema_id="cl5it7ktp00i5ci6gf80b1ysd",
                    extra={"uuid": "f963be22-227b-4efe-9be4-2738ed822216"},
                    value=Point(x=50.0, y=50.0),
                    frame=5,
                    keyframe=True,
                    segment_index=1,
                ),
                VideoObjectAnnotation(
                    feature_schema_id="cl5it7ktp00i5ci6gf80b1ysd",
                    extra={"uuid": "f963be22-227b-4efe-9be4-2738ed822216"},
                    value=Point(x=10.0, y=50.0),
                    frame=10,
                    keyframe=True,
                    segment_index=1,
                ),
                VideoObjectAnnotation(
                    feature_schema_id="cl5iw0roz00lwci6g5jni62vs",
                    extra={"uuid": "13b2ee0e-2355-4336-8b83-d74d09e3b1e7"},
                    value=Rectangle(
                        start=Point(x=5.0, y=10.0),
                        end=Point(x=155.0, y=110.0),
                    ),
                    frame=1,
                    keyframe=True,
                    segment_index=0,
                ),
                VideoObjectAnnotation(
                    feature_schema_id="cl5iw0roz00lwci6g5jni62vs",
                    extra={"uuid": "13b2ee0e-2355-4336-8b83-d74d09e3b1e7"},
                    value=Rectangle(
                        start=Point(x=5.0, y=30.0),
                        end=Point(x=155.0, y=80.0),
                    ),
                    frame=5,
                    keyframe=True,
                    segment_index=0,
                ),
                VideoObjectAnnotation(
                    feature_schema_id="cl5iw0roz00lwci6g5jni62vs",
                    extra={"uuid": "13b2ee0e-2355-4336-8b83-d74d09e3b1e7"},
                    value=Rectangle(
                        start=Point(x=200.0, y=300.0),
                        end=Point(x=350.0, y=700.0),
                    ),
                    frame=10,
                    keyframe=True,
                    segment_index=1,
                ),
            ],
        )
    ]

    res = list(NDJsonConverter.serialize(labels))

    data = sorted(data, key=itemgetter("uuid"))
    res = sorted(res, key=itemgetter("uuid"))
    assert data == res


def test_video_name_only():
    with open(
        "tests/data/assets/ndjson/video_import_name_only.json", "r"
    ) as file:
        data = json.load(file)
    labels = [
        Label(
            data=GenericDataRowData(uid="ckrb1sf1i1g7i0ybcdc6oc8ct"),
            annotations=[
                VideoClassificationAnnotation(
                    name="question 1",
                    extra={"uuid": "f6879f59-d2b5-49c2-aceb-d9e8dc478673"},
                    value=Radio(
                        answer=ClassificationAnswer(
                            name="answer 1",
                        ),
                    ),
                    frame=30,
                ),
                VideoClassificationAnnotation(
                    name="question 1",
                    extra={"uuid": "f6879f59-d2b5-49c2-aceb-d9e8dc478673"},
                    value=Radio(
                        answer=ClassificationAnswer(
                            name="answer 1",
                        ),
                    ),
                    frame=31,
                ),
                VideoClassificationAnnotation(
                    name="question 1",
                    extra={"uuid": "f6879f59-d2b5-49c2-aceb-d9e8dc478673"},
                    value=Radio(
                        answer=ClassificationAnswer(
                            name="answer 1",
                        ),
                    ),
                    frame=32,
                ),
                VideoClassificationAnnotation(
                    name="question 1",
                    extra={"uuid": "f6879f59-d2b5-49c2-aceb-d9e8dc478673"},
                    value=Radio(
                        answer=ClassificationAnswer(
                            name="answer 1",
                        ),
                    ),
                    frame=33,
                ),
                VideoClassificationAnnotation(
                    name="question 1",
                    extra={"uuid": "f6879f59-d2b5-49c2-aceb-d9e8dc478673"},
                    value=Radio(
                        answer=ClassificationAnswer(
                            name="answer 1",
                        ),
                    ),
                    frame=34,
                ),
                VideoClassificationAnnotation(
                    name="question 1",
                    extra={"uuid": "f6879f59-d2b5-49c2-aceb-d9e8dc478673"},
                    value=Radio(
                        answer=ClassificationAnswer(
                            name="answer 1",
                        ),
                    ),
                    frame=35,
                ),
                VideoClassificationAnnotation(
                    name="question 1",
                    extra={"uuid": "f6879f59-d2b5-49c2-aceb-d9e8dc478673"},
                    value=Radio(
                        answer=ClassificationAnswer(
                            name="answer 1",
                        ),
                    ),
                    frame=50,
                ),
                VideoClassificationAnnotation(
                    name="question 1",
                    extra={"uuid": "f6879f59-d2b5-49c2-aceb-d9e8dc478673"},
                    value=Radio(
                        answer=ClassificationAnswer(
                            name="answer 1",
                        ),
                    ),
                    frame=51,
                ),
                VideoClassificationAnnotation(
                    name="question 2",
                    extra={"uuid": "d009925d-91a3-4f67-abd9-753453f5a584"},
                    value=Checklist(
                        answer=[
                            ClassificationAnswer(
                                name="answer 2",
                            )
                        ],
                    ),
                    frame=0,
                ),
                VideoClassificationAnnotation(
                    name="question 2",
                    extra={"uuid": "d009925d-91a3-4f67-abd9-753453f5a584"},
                    value=Checklist(
                        answer=[
                            ClassificationAnswer(
                                name="answer 2",
                            )
                        ],
                    ),
                    frame=1,
                ),
                VideoClassificationAnnotation(
                    name="question 2",
                    extra={"uuid": "d009925d-91a3-4f67-abd9-753453f5a584"},
                    value=Checklist(
                        answer=[
                            ClassificationAnswer(
                                name="answer 2",
                            )
                        ],
                    ),
                    frame=2,
                ),
                VideoClassificationAnnotation(
                    name="question 2",
                    extra={"uuid": "d009925d-91a3-4f67-abd9-753453f5a584"},
                    value=Checklist(
                        answer=[
                            ClassificationAnswer(
                                name="answer 2",
                            )
                        ],
                    ),
                    frame=3,
                ),
                VideoClassificationAnnotation(
                    name="question 2",
                    extra={"uuid": "d009925d-91a3-4f67-abd9-753453f5a584"},
                    value=Checklist(
                        answer=[
                            ClassificationAnswer(
                                name="answer 2",
                            )
                        ],
                    ),
                    frame=4,
                ),
                VideoClassificationAnnotation(
                    name="question 2",
                    extra={"uuid": "d009925d-91a3-4f67-abd9-753453f5a584"},
                    value=Checklist(
                        answer=[
                            ClassificationAnswer(
                                name="answer 2",
                            )
                        ],
                    ),
                    frame=5,
                ),
                ClassificationAnnotation(
                    name="question 3",
                    extra={"uuid": "e5f32456-bd67-4520-8d3b-cbeb2204bad3"},
                    value=Text(answer="a value"),
                ),
                VideoObjectAnnotation(
                    name="segment 1",
                    extra={"uuid": "6f7c835a-0139-4896-b73f-66a6baa89e94"},
                    value=Line(
                        points=[
                            Point(x=10.0, y=10.0),
                            Point(x=100.0, y=100.0),
                            Point(x=50.0, y=30.0),
                        ],
                    ),
                    frame=1,
                    keyframe=True,
                    segment_index=0,
                ),
                VideoObjectAnnotation(
                    name="segment 1",
                    extra={"uuid": "6f7c835a-0139-4896-b73f-66a6baa89e94"},
                    value=Line(
                        points=[
                            Point(x=15.0, y=10.0),
                            Point(x=50.0, y=100.0),
                            Point(x=50.0, y=30.0),
                        ],
                    ),
                    frame=5,
                    keyframe=True,
                    segment_index=0,
                ),
                VideoObjectAnnotation(
                    name="segment 1",
                    extra={"uuid": "6f7c835a-0139-4896-b73f-66a6baa89e94"},
                    value=Line(
                        points=[
                            Point(x=100.0, y=10.0),
                            Point(x=50.0, y=100.0),
                            Point(x=50.0, y=30.0),
                        ],
                    ),
                    frame=8,
                    keyframe=True,
                    segment_index=1,
                ),
                VideoObjectAnnotation(
                    name="segment 2",
                    extra={"uuid": "f963be22-227b-4efe-9be4-2738ed822216"},
                    value=Point(x=10.0, y=10.0),
                    frame=1,
                    keyframe=True,
                    segment_index=0,
                ),
                VideoObjectAnnotation(
                    name="segment 2",
                    extra={"uuid": "f963be22-227b-4efe-9be4-2738ed822216"},
                    value=Point(x=50.0, y=50.0),
                    frame=5,
                    keyframe=True,
                    segment_index=1,
                ),
                VideoObjectAnnotation(
                    name="segment 2",
                    extra={"uuid": "f963be22-227b-4efe-9be4-2738ed822216"},
                    value=Point(x=10.0, y=50.0),
                    frame=10,
                    keyframe=True,
                    segment_index=1,
                ),
                VideoObjectAnnotation(
                    name="segment 3",
                    extra={"uuid": "13b2ee0e-2355-4336-8b83-d74d09e3b1e7"},
                    value=Rectangle(
                        start=Point(x=5.0, y=10.0),
                        end=Point(x=155.0, y=110.0),
                    ),
                    frame=1,
                    keyframe=True,
                    segment_index=0,
                ),
                VideoObjectAnnotation(
                    name="segment 3",
                    extra={"uuid": "13b2ee0e-2355-4336-8b83-d74d09e3b1e7"},
                    value=Rectangle(
                        start=Point(x=5.0, y=30.0),
                        end=Point(x=155.0, y=80.0),
                    ),
                    frame=5,
                    keyframe=True,
                    segment_index=0,
                ),
                VideoObjectAnnotation(
                    name="segment 3",
                    extra={"uuid": "13b2ee0e-2355-4336-8b83-d74d09e3b1e7"},
                    value=Rectangle(
                        start=Point(x=200.0, y=300.0),
                        end=Point(x=350.0, y=700.0),
                    ),
                    frame=10,
                    keyframe=True,
                    segment_index=1,
                ),
            ],
        )
    ]
    res = list(NDJsonConverter.serialize(labels))
    data = sorted(data, key=itemgetter("uuid"))
    res = sorted(res, key=itemgetter("uuid"))

    assert data == res


def test_video_classification_global_subclassifications():
    label = Label(
        data=GenericDataRowData(
            global_key="sample-video-4.mp4",
        ),
        annotations=[
            ClassificationAnnotation(
                name="radio_question_nested",
                value=Radio(
                    answer=ClassificationAnswer(name="first_radio_question")
                ),
            ),
            ClassificationAnnotation(
                name="nested_checklist_question",
                value=Checklist(
                    answer=[
                        ClassificationAnswer(
                            name="first_checklist_answer",
                            classifications=[
                                ClassificationAnnotation(
                                    name="sub_checklist_question",
                                    value=Radio(
                                        answer=ClassificationAnswer(
                                            name="first_sub_checklist_answer"
                                        )
                                    ),
                                )
                            ],
                        )
                    ],
                ),
            ),
        ],
    )

    expected_first_annotation = {
        "name": "radio_question_nested",
        "answer": {"name": "first_radio_question"},
        "dataRow": {"globalKey": "sample-video-4.mp4"},
    }

    expected_second_annotation = {
        "name": "nested_checklist_question",
        "answer": [
            {
                "name": "first_checklist_answer",
                "classifications": [
                    {
                        "name": "sub_checklist_question",
                        "answer": {"name": "first_sub_checklist_answer"},
                    }
                ],
            }
        ],
        "dataRow": {"globalKey": "sample-video-4.mp4"},
    }

    serialized = NDJsonConverter.serialize([label])
    res = [x for x in serialized]
    for annotations in res:
        annotations.pop("uuid")
    assert res == [expected_first_annotation, expected_second_annotation]


def test_video_classification_nesting_bbox():
    bbox_annotation = [
        VideoObjectAnnotation(
            name="bbox_video",
            keyframe=True,
            frame=13,
            segment_index=0,
            value=Rectangle(
                start=Point(x=146.0, y=98.0),  # Top left
                end=Point(x=382.0, y=341.0),  # Bottom right
            ),
            classifications=[
                ClassificationAnnotation(
                    name="radio_question_nested",
                    value=Radio(
                        answer=ClassificationAnswer(
                            name="first_radio_question",
                            classifications=[
                                ClassificationAnnotation(
                                    name="sub_question_radio",
                                    value=Checklist(
                                        answer=[
                                            ClassificationAnswer(
                                                name="sub_answer"
                                            )
                                        ]
                                    ),
                                )
                            ],
                        )
                    ),
                )
            ],
        ),
        VideoObjectAnnotation(
            name="bbox_video",
            keyframe=True,
            frame=15,
            segment_index=0,
            value=Rectangle(
                start=Point(x=146.0, y=98.0),  # Top left
                end=Point(x=382.0, y=341.0),  # Bottom right,
            ),
            classifications=[
                ClassificationAnnotation(
                    name="nested_checklist_question",
                    value=Checklist(
                        answer=[
                            ClassificationAnswer(
                                name="first_checklist_answer",
                                classifications=[
                                    ClassificationAnnotation(
                                        name="sub_checklist_question",
                                        value=Radio(
                                            answer=ClassificationAnswer(
                                                name="first_sub_checklist_answer"
                                            )
                                        ),
                                    )
                                ],
                            )
                        ]
                    ),
                )
            ],
        ),
        VideoObjectAnnotation(
            name="bbox_video",
            keyframe=True,
            frame=19,
            segment_index=0,
            value=Rectangle(
                start=Point(x=146.0, y=98.0),  # Top left
                end=Point(x=382.0, y=341.0),  # Bottom right
            ),
        ),
    ]
    expected = [
        {
            "dataRow": {"globalKey": "sample-video-4.mp4"},
            "name": "bbox_video",
            "classifications": [],
            "segments": [
                {
                    "keyframes": [
                        {
                            "frame": 13,
                            "bbox": {
                                "top": 98.0,
                                "left": 146.0,
                                "height": 243.0,
                                "width": 236.0,
                            },
                            "classifications": [
                                {
                                    "name": "radio_question_nested",
                                    "answer": {
                                        "name": "first_radio_question",
                                        "classifications": [
                                            {
                                                "name": "sub_question_radio",
                                                "answer": [
                                                    {"name": "sub_answer"}
                                                ],
                                            }
                                        ],
                                    },
                                }
                            ],
                        },
                        {
                            "frame": 15,
                            "bbox": {
                                "top": 98.0,
                                "left": 146.0,
                                "height": 243.0,
                                "width": 236.0,
                            },
                            "classifications": [
                                {
                                    "name": "nested_checklist_question",
                                    "answer": [
                                        {
                                            "name": "first_checklist_answer",
                                            "classifications": [
                                                {
                                                    "name": "sub_checklist_question",
                                                    "answer": {
                                                        "name": "first_sub_checklist_answer"
                                                    },
                                                }
                                            ],
                                        }
                                    ],
                                }
                            ],
                        },
                        {
                            "frame": 19,
                            "bbox": {
                                "top": 98.0,
                                "left": 146.0,
                                "height": 243.0,
                                "width": 236.0,
                            },
                            "classifications": [],
                        },
                    ]
                }
            ],
        }
    ]

    label = Label(
        data=GenericDataRowData(
            global_key="sample-video-4.mp4",
        ),
        annotations=bbox_annotation,
    )

    serialized = NDJsonConverter.serialize([label])
    res = [x for x in serialized]
    assert res == expected


def test_video_classification_point():
    bbox_annotation = [
        VideoObjectAnnotation(
            name="bbox_video",
            keyframe=True,
            frame=13,
            segment_index=0,
            value=Point(x=46.0, y=8.0),
            classifications=[
                ClassificationAnnotation(
                    name="radio_question_nested",
                    value=Radio(
                        answer=ClassificationAnswer(
                            name="first_radio_question",
                            classifications=[
                                ClassificationAnnotation(
                                    name="sub_question_radio",
                                    value=Checklist(
                                        answer=[
                                            ClassificationAnswer(
                                                name="sub_answer"
                                            )
                                        ]
                                    ),
                                )
                            ],
                        )
                    ),
                )
            ],
        ),
        VideoObjectAnnotation(
            name="bbox_video",
            keyframe=True,
            frame=15,
            segment_index=0,
            value=Point(x=56.0, y=18.0),
            classifications=[
                ClassificationAnnotation(
                    name="nested_checklist_question",
                    value=Checklist(
                        answer=[
                            ClassificationAnswer(
                                name="first_checklist_answer",
                                classifications=[
                                    ClassificationAnnotation(
                                        name="sub_checklist_question",
                                        value=Radio(
                                            answer=ClassificationAnswer(
                                                name="first_sub_checklist_answer"
                                            )
                                        ),
                                    )
                                ],
                            )
                        ]
                    ),
                )
            ],
        ),
        VideoObjectAnnotation(
            name="bbox_video",
            keyframe=True,
            frame=19,
            segment_index=0,
            value=Point(x=66.0, y=28.0),
        ),
    ]
    expected = [
        {
            "dataRow": {"globalKey": "sample-video-4.mp4"},
            "name": "bbox_video",
            "classifications": [],
            "segments": [
                {
                    "keyframes": [
                        {
                            "frame": 13,
                            "point": {
                                "x": 46.0,
                                "y": 8.0,
                            },
                            "classifications": [
                                {
                                    "name": "radio_question_nested",
                                    "answer": {
                                        "name": "first_radio_question",
                                        "classifications": [
                                            {
                                                "name": "sub_question_radio",
                                                "answer": [
                                                    {"name": "sub_answer"}
                                                ],
                                            }
                                        ],
                                    },
                                }
                            ],
                        },
                        {
                            "frame": 15,
                            "point": {
                                "x": 56.0,
                                "y": 18.0,
                            },
                            "classifications": [
                                {
                                    "name": "nested_checklist_question",
                                    "answer": [
                                        {
                                            "name": "first_checklist_answer",
                                            "classifications": [
                                                {
                                                    "name": "sub_checklist_question",
                                                    "answer": {
                                                        "name": "first_sub_checklist_answer"
                                                    },
                                                }
                                            ],
                                        }
                                    ],
                                }
                            ],
                        },
                        {
                            "frame": 19,
                            "point": {
                                "x": 66.0,
                                "y": 28.0,
                            },
                            "classifications": [],
                        },
                    ]
                }
            ],
        }
    ]

    label = Label(
        data=GenericDataRowData(
            global_key="sample-video-4.mp4",
        ),
        annotations=bbox_annotation,
    )

    serialized = NDJsonConverter.serialize([label])
    res = [x for x in serialized]
    assert res == expected


def test_video_classification_frameline():
    bbox_annotation = [
        VideoObjectAnnotation(
            name="bbox_video",
            keyframe=True,
            frame=13,
            segment_index=0,
            value=Line(points=[Point(x=8, y=10), Point(x=10, y=9)]),
            classifications=[
                ClassificationAnnotation(
                    name="radio_question_nested",
                    value=Radio(
                        answer=ClassificationAnswer(
                            name="first_radio_question",
                            classifications=[
                                ClassificationAnnotation(
                                    name="sub_question_radio",
                                    value=Checklist(
                                        answer=[
                                            ClassificationAnswer(
                                                name="sub_answer"
                                            )
                                        ]
                                    ),
                                )
                            ],
                        )
                    ),
                )
            ],
        ),
        VideoObjectAnnotation(
            name="bbox_video",
            keyframe=True,
            frame=15,
            segment_index=0,
            value=Line(points=[Point(x=18, y=20), Point(x=20, y=19)]),
            classifications=[
                ClassificationAnnotation(
                    name="nested_checklist_question",
                    value=Checklist(
                        answer=[
                            ClassificationAnswer(
                                name="first_checklist_answer",
                                classifications=[
                                    ClassificationAnnotation(
                                        name="sub_checklist_question",
                                        value=Radio(
                                            answer=ClassificationAnswer(
                                                name="first_sub_checklist_answer"
                                            )
                                        ),
                                    )
                                ],
                            )
                        ]
                    ),
                )
            ],
        ),
        VideoObjectAnnotation(
            name="bbox_video",
            keyframe=True,
            frame=19,
            segment_index=0,
            value=Line(points=[Point(x=28, y=30), Point(x=30, y=29)]),
        ),
    ]
    expected = [
        {
            "dataRow": {"globalKey": "sample-video-4.mp4"},
            "name": "bbox_video",
            "classifications": [],
            "segments": [
                {
                    "keyframes": [
                        {
                            "frame": 13,
                            "line": [
                                {
                                    "x": 8.0,
                                    "y": 10.0,
                                },
                                {
                                    "x": 10.0,
                                    "y": 9.0,
                                },
                            ],
                            "classifications": [
                                {
                                    "name": "radio_question_nested",
                                    "answer": {
                                        "name": "first_radio_question",
                                        "classifications": [
                                            {
                                                "name": "sub_question_radio",
                                                "answer": [
                                                    {"name": "sub_answer"}
                                                ],
                                            }
                                        ],
                                    },
                                }
                            ],
                        },
                        {
                            "frame": 15,
                            "line": [
                                {
                                    "x": 18.0,
                                    "y": 20.0,
                                },
                                {
                                    "x": 20.0,
                                    "y": 19.0,
                                },
                            ],
                            "classifications": [
                                {
                                    "name": "nested_checklist_question",
                                    "answer": [
                                        {
                                            "name": "first_checklist_answer",
                                            "classifications": [
                                                {
                                                    "name": "sub_checklist_question",
                                                    "answer": {
                                                        "name": "first_sub_checklist_answer"
                                                    },
                                                }
                                            ],
                                        }
                                    ],
                                }
                            ],
                        },
                        {
                            "frame": 19,
                            "line": [
                                {
                                    "x": 28.0,
                                    "y": 30.0,
                                },
                                {
                                    "x": 30.0,
                                    "y": 29.0,
                                },
                            ],
                            "classifications": [],
                        },
                    ]
                }
            ],
        }
    ]

    label = Label(
        data=GenericDataRowData(
            global_key="sample-video-4.mp4",
        ),
        annotations=bbox_annotation,
    )
    serialized = NDJsonConverter.serialize([label])
    res = [x for x in serialized]
    assert res == expected


[
    {
        "answer": "a value",
        "dataRow": {"id": "ckrb1sf1i1g7i0ybcdc6oc8ct"},
        "schemaId": "ckrb1sfkn099c0y910wbo0p1a",
        "uuid": "90e2ecf7-c19c-47e6-8cdb-8867e1b9d88c",
    },
    {
        "answer": {"schemaId": "ckrb1sfl8099g0y91cxbd5ftb"},
        "dataRow": {"id": "ckrb1sf1i1g7i0ybcdc6oc8ct"},
        "frames": [{"end": 35, "start": 30}, {"end": 51, "start": 50}],
        "schemaId": "ckrb1sfjx099a0y914hl319ie",
        "uuid": "f6879f59-d2b5-49c2-aceb-d9e8dc478673",
    },
    {
        "answer": [{"schemaId": "ckrb1sfl8099e0y919v260awv"}],
        "dataRow": {"id": "ckrb1sf1i1g7i0ybcdc6oc8ct"},
        "frames": [{"end": 5, "start": 0}],
        "schemaId": "ckrb1sfkn099c0y910wbo0p1a",
        "uuid": "d009925d-91a3-4f67-abd9-753453f5a584",
    },
    {
        "classifications": [],
        "dataRow": {"id": "ckrb1sf1i1g7i0ybcdc6oc8ct"},
        "schemaId": "cl5islwg200gfci6g0oitaypu",
        "segments": [
            {
                "keyframes": [
                    {
                        "classifications": [],
                        "frame": 1,
                        "line": [
                            {"x": 10.0, "y": 10.0},
                            {"x": 100.0, "y": 100.0},
                            {"x": 50.0, "y": 30.0},
                        ],
                    },
                    {
                        "classifications": [],
                        "frame": 5,
                        "line": [
                            {"x": 15.0, "y": 10.0},
                            {"x": 50.0, "y": 100.0},
                            {"x": 50.0, "y": 30.0},
                        ],
                    },
                ]
            },
            {
                "keyframes": [
                    {
                        "classifications": [],
                        "frame": 8,
                        "line": [
                            {"x": 100.0, "y": 10.0},
                            {"x": 50.0, "y": 100.0},
                            {"x": 50.0, "y": 30.0},
                        ],
                    }
                ]
            },
        ],
        "uuid": "6f7c835a-0139-4896-b73f-66a6baa89e94",
    },
    {
        "classifications": [],
        "dataRow": {"id": "ckrb1sf1i1g7i0ybcdc6oc8ct"},
        "schemaId": "cl5it7ktp00i5ci6gf80b1ysd",
        "segments": [
            {
                "keyframes": [
                    {
                        "classifications": [],
                        "frame": 1,
                        "point": {"x": 10.0, "y": 10.0},
                    }
                ]
            },
            {
                "keyframes": [
                    {
                        "classifications": [],
                        "frame": 5,
                        "point": {"x": 50.0, "y": 50.0},
                    },
                    {
                        "classifications": [],
                        "frame": 10,
                        "point": {"x": 10.0, "y": 50.0},
                    },
                ]
            },
        ],
        "uuid": "f963be22-227b-4efe-9be4-2738ed822216",
    },
    {
        "classifications": [],
        "dataRow": {"id": "ckrb1sf1i1g7i0ybcdc6oc8ct"},
        "schemaId": "cl5iw0roz00lwci6g5jni62vs",
        "segments": [
            {
                "keyframes": [
                    {
                        "bbox": {
                            "height": 100.0,
                            "left": 5.0,
                            "top": 10.0,
                            "width": 150.0,
                        },
                        "classifications": [],
                        "frame": 1,
                    },
                    {
                        "bbox": {
                            "height": 50.0,
                            "left": 5.0,
                            "top": 30.0,
                            "width": 150.0,
                        },
                        "classifications": [],
                        "frame": 5,
                    },
                ]
            },
            {
                "keyframes": [
                    {
                        "bbox": {
                            "height": 400.0,
                            "left": 200.0,
                            "top": 300.0,
                            "width": 150.0,
                        },
                        "classifications": [],
                        "frame": 10,
                    }
                ]
            },
        ],
        "uuid": "13b2ee0e-2355-4336-8b83-d74d09e3b1e7",
    },
]

[
    {
        "answer": {"schemaId": "ckrb1sfl8099g0y91cxbd5ftb"},
        "schemaId": "ckrb1sfjx099a0y914hl319ie",
        "dataRow": {"id": "ckrb1sf1i1g7i0ybcdc6oc8ct"},
        "uuid": "f6879f59-d2b5-49c2-aceb-d9e8dc478673",
        "frames": [{"start": 30, "end": 35}, {"start": 50, "end": 51}],
    },
    {
        "answer": [{"schemaId": "ckrb1sfl8099e0y919v260awv"}],
        "schemaId": "ckrb1sfkn099c0y910wbo0p1a",
        "dataRow": {"id": "ckrb1sf1i1g7i0ybcdc6oc8ct"},
        "uuid": "d009925d-91a3-4f67-abd9-753453f5a584",
        "frames": [{"start": 0, "end": 5}],
    },
    {
        "answer": "a value",
        "schemaId": "ckrb1sfkn099c0y910wbo0p1a",
        "dataRow": {"id": "ckrb1sf1i1g7i0ybcdc6oc8ct"},
        "uuid": "90e2ecf7-c19c-47e6-8cdb-8867e1b9d88c",
    },
    {
        "classifications": [],
        "schemaId": "cl5islwg200gfci6g0oitaypu",
        "dataRow": {"id": "ckrb1sf1i1g7i0ybcdc6oc8ct"},
        "uuid": "6f7c835a-0139-4896-b73f-66a6baa89e94",
        "segments": [
            {
                "keyframes": [
                    {
                        "frame": 1,
                        "line": [
                            {"x": 10.0, "y": 10.0},
                            {"x": 100.0, "y": 100.0},
                            {"x": 50.0, "y": 30.0},
                        ],
                        "classifications": [],
                    },
                    {
                        "frame": 5,
                        "line": [
                            {"x": 15.0, "y": 10.0},
                            {"x": 50.0, "y": 100.0},
                            {"x": 50.0, "y": 30.0},
                        ],
                        "classifications": [],
                    },
                ]
            },
            {
                "keyframes": [
                    {
                        "frame": 8,
                        "line": [
                            {"x": 100.0, "y": 10.0},
                            {"x": 50.0, "y": 100.0},
                            {"x": 50.0, "y": 30.0},
                        ],
                        "classifications": [],
                    }
                ]
            },
        ],
    },
    {
        "classifications": [],
        "schemaId": "cl5it7ktp00i5ci6gf80b1ysd",
        "dataRow": {"id": "ckrb1sf1i1g7i0ybcdc6oc8ct"},
        "uuid": "f963be22-227b-4efe-9be4-2738ed822216",
        "segments": [
            {
                "keyframes": [
                    {
                        "frame": 1,
                        "point": {"x": 10.0, "y": 10.0},
                        "classifications": [],
                    }
                ]
            },
            {
                "keyframes": [
                    {
                        "frame": 5,
                        "point": {"x": 50.0, "y": 50.0},
                        "classifications": [],
                    },
                    {
                        "frame": 10,
                        "point": {"x": 10.0, "y": 50.0},
                        "classifications": [],
                    },
                ]
            },
        ],
    },
    {
        "classifications": [],
        "schemaId": "cl5iw0roz00lwci6g5jni62vs",
        "dataRow": {"id": "ckrb1sf1i1g7i0ybcdc6oc8ct"},
        "uuid": "13b2ee0e-2355-4336-8b83-d74d09e3b1e7",
        "segments": [
            {
                "keyframes": [
                    {
                        "frame": 1,
                        "bbox": {
                            "top": 10.0,
                            "left": 5.0,
                            "height": 100.0,
                            "width": 150.0,
                        },
                        "classifications": [],
                    },
                    {
                        "frame": 5,
                        "bbox": {
                            "top": 30.0,
                            "left": 5.0,
                            "height": 50.0,
                            "width": 150.0,
                        },
                        "classifications": [],
                    },
                ]
            },
            {
                "keyframes": [
                    {
                        "frame": 10,
                        "bbox": {
                            "top": 300.0,
                            "left": 200.0,
                            "height": 400.0,
                            "width": 150.0,
                        },
                        "classifications": [],
                    }
                ]
            },
        ],
    },
]



import pytest
import uuid
from typing import Tuple
from labelbox.schema.export_task import ExportTask, StreamType

from labelbox.schema.media_type import MediaType
from labelbox import Project, Dataset
from labelbox.schema.data_row import DataRow
from labelbox.schema.label import Label
from labelbox import UniqueIds

IMAGE_URL = "https://storage.googleapis.com/lb-artifacts-testing-public/sdk_integration_test/potato.jpeg"


class TestExportProject:
    @pytest.fixture
    def project_export(self):
        def _project_export(project, task_name, filters=None, params=None):
            export_task = project.export(
                task_name=task_name,
                filters=filters,
                params=params,
            )
            export_task.wait_till_done()

            assert export_task.status == "COMPLETE"
            assert isinstance(export_task, ExportTask)
            return export_task

        return _project_export

    def test_export(
        self,
        client,
        configured_project_with_label,
        wait_for_data_row_processing,
        project_export,
    ):
        project, dataset, data_row, label = configured_project_with_label
        data_row = wait_for_data_row_processing(client, data_row)
        label_id = label.uid
        task_name = "TestExportProject:test_export"
        params = {
            "include_performance_details": True,
            "include_labels": True,
            "media_type_override": MediaType.Image,
            "project_details": True,
            "data_row_details": True,
        }
        export_task = project_export(project, task_name, params=params)
        assert export_task.has_result()
        assert export_task.has_errors() is False
        assert (
            export_task.get_total_file_size(stream_type=StreamType.RESULT) > 0
        )
        assert export_task.get_total_lines(stream_type=StreamType.RESULT) > 0

        for data in export_task.get_buffered_stream():
            obj = data.json
            task_media_attributes = obj["media_attributes"]
            task_project = obj["projects"][project.uid]
            task_project_label_ids_set = set(
                map(lambda prediction: prediction["id"], task_project["labels"])
            )
            task_project_details = task_project["project_details"]
            task_data_row = obj["data_row"]
            task_data_row_details = task_data_row["details"]

            assert label_id in task_project_label_ids_set
            # data row
            assert task_data_row["id"] == data_row.uid
            assert task_data_row["external_id"] == data_row.external_id
            assert task_data_row["row_data"] == data_row.row_data

            # data row details
            assert task_data_row_details["dataset_id"] == dataset.uid
            assert task_data_row_details["dataset_name"] == dataset.name

            assert task_data_row_details["last_activity_at"] is not None
            assert task_data_row_details["created_by"] is not None

            # media attributes
            assert (
                task_media_attributes["mime_type"]
                == data_row.media_attributes["mimeType"]
            )

            # project name and details
            assert task_project["name"] == project.name
            batch = next(project.batches())
            assert task_project_details["batch_id"] == batch.uid
            assert task_project_details["batch_name"] == batch.name
            assert task_project_details["priority"] is not None
            assert (
                task_project_details["consensus_expected_label_count"]
                is not None
            )
            assert task_project_details["workflow_history"] is not None

            # label details
            assert task_project["labels"][0]["id"] == label_id

    def test_with_date_filters(
        self,
        client,
        configured_project_with_label,
        wait_for_data_row_processing,
        project_export,
    ):
        project, _, data_row, label = configured_project_with_label
        data_row = wait_for_data_row_processing(client, data_row)
        label_id = label.uid
        task_name = "TestExportProject:test_with_date_filters"
        filters = {
            "last_activity_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
            "label_created_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
            "task_queue_status": "InReview",
        }
        include_performance_details = True
        params = {
            "performance_details": include_performance_details,
            "include_labels": True,
            "project_details": True,
            "media_type_override": MediaType.Image,
        }
        task_queues = project.task_queues()
        review_queue = next(
            tq for tq in task_queues if tq.queue_type == "MANUAL_REVIEW_QUEUE"
        )
        project.move_data_rows_to_task_queue(
            UniqueIds([data_row.uid]), review_queue.uid
        )
        export_task = project_export(
            project, task_name, filters=filters, params=params
        )
        assert export_task.has_result()
        assert export_task.has_errors() is False
        assert (
            export_task.get_total_file_size(stream_type=StreamType.RESULT) > 0
        )
        assert export_task.get_total_lines(stream_type=StreamType.RESULT) > 0

        for data in export_task.get_buffered_stream():
            obj = data.json
            task_project = obj["projects"][project.uid]
            task_project_label_ids_set = set(
                map(lambda prediction: prediction["id"], task_project["labels"])
            )
            assert label_id in task_project_label_ids_set
            assert (
                task_project["project_details"]["workflow_status"]
                == "IN_REVIEW"
            )

    def test_with_iso_date_filters(
        self,
        client,
        configured_project_with_label,
        wait_for_data_row_processing,
        project_export,
    ):
        project, _, data_row, label = configured_project_with_label
        data_row = wait_for_data_row_processing(client, data_row)
        label_id = label.uid
        task_name = "TestExportProject:test_with_iso_date_filters"
        filters = {
            "last_activity_at": [
                "2000-01-01T00:00:00+0230",
                "2050-01-01T00:00:00+0230",
            ],
            "label_created_at": [
                "2000-01-01T00:00:00+0230",
                "2050-01-01T00:00:00+0230",
            ],
        }
        export_task = project_export(project, task_name, filters=filters)
        assert export_task.has_result()
        assert export_task.has_errors() is False
        assert (
            export_task.get_total_file_size(stream_type=StreamType.RESULT) > 0
        )
        assert export_task.get_total_lines(stream_type=StreamType.RESULT) > 0
        assert (
            label_id
            == list(export_task.get_buffered_stream())[0].json["projects"][
                project.uid
            ]["labels"][0]["id"]
        )

    def test_with_iso_date_filters_no_start_date(
        self,
        client,
        configured_project_with_label,
        wait_for_data_row_processing,
        project_export,
    ):
        project, _, data_row, label = configured_project_with_label
        data_row = wait_for_data_row_processing(client, data_row)
        label_id = label.uid
        task_name = "TestExportProject:test_with_iso_date_filters_no_start_date"
        filters = {"last_activity_at": [None, "2050-01-01T00:00:00+0230"]}
        export_task = project_export(project, task_name, filters=filters)
        assert export_task.has_result()
        assert export_task.has_errors() is False
        assert (
            export_task.get_total_file_size(stream_type=StreamType.RESULT) > 0
        )
        assert export_task.get_total_lines(stream_type=StreamType.RESULT) > 0
        assert (
            label_id
            == list(export_task.get_buffered_stream())[0].json["projects"][
                project.uid
            ]["labels"][0]["id"]
        )

    def test_with_iso_date_filters_and_future_start_date(
        self,
        client,
        configured_project_with_label,
        wait_for_data_row_processing,
        project_export,
    ):
        project, _, data_row, _label = configured_project_with_label
        data_row = wait_for_data_row_processing(client, data_row)
        task_name = (
            "TestExportProject:test_with_iso_date_filters_and_future_start_date"
        )
        filters = {"label_created_at": ["2050-01-01T00:00:00+0230", None]}
        export_task = project_export(project, task_name, filters=filters)
        assert export_task.has_result() is False
        assert export_task.has_errors() is False
        assert (
            export_task.get_total_file_size(stream_type=StreamType.RESULT)
            is None
        )
        assert (
            export_task.get_total_lines(stream_type=StreamType.RESULT) is None
        )

    @pytest.mark.parametrize("data_rows", [3], indirect=True)
    def test_with_data_row_filter(
        self, configured_batch_project_with_multiple_datarows, project_export
    ):
        project, _, data_rows = configured_batch_project_with_multiple_datarows
        datarow_filter_size = 2
        expected_data_row_ids = [dr.uid for dr in data_rows][
            :datarow_filter_size
        ]
        task_name = "TestExportProject:test_with_data_row_filter"
        filters = {
            "last_activity_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
            "label_created_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
            "data_row_ids": expected_data_row_ids,
        }
        params = {
            "data_row_details": True,
            "media_type_override": MediaType.Image,
        }
        export_task = project_export(
            project, task_name, filters=filters, params=params
        )
        assert export_task.has_result()
        assert export_task.has_errors() is False
        assert (
            export_task.get_total_file_size(stream_type=StreamType.RESULT) > 0
        )
        # only 2 datarows should be exported
        assert (
            export_task.get_total_lines(stream_type=StreamType.RESULT)
            == datarow_filter_size
        )
        data_row_ids = list(
            map(
                lambda x: x.json["data_row"]["id"],
                export_task.get_buffered_stream(),
            )
        )
        assert data_row_ids.sort() == expected_data_row_ids.sort()

    @pytest.mark.parametrize("data_rows", [3], indirect=True)
    def test_with_global_key_filter(
        self, configured_batch_project_with_multiple_datarows, project_export
    ):
        project, _, data_rows = configured_batch_project_with_multiple_datarows
        datarow_filter_size = 2
        expected_global_keys = [dr.global_key for dr in data_rows][
            :datarow_filter_size
        ]
        task_name = "TestExportProject:test_with_global_key_filter"
        filters = {
            "last_activity_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
            "label_created_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
            "global_keys": expected_global_keys,
        }
        params = {
            "data_row_details": True,
            "media_type_override": MediaType.Image,
        }
        export_task = project_export(
            project, task_name, filters=filters, params=params
        )
        assert export_task.has_result()
        assert export_task.has_errors() is False
        assert (
            export_task.get_total_file_size(stream_type=StreamType.RESULT) > 0
        )
        # only 2 datarows should be exported
        assert (
            export_task.get_total_lines(stream_type=StreamType.RESULT)
            == datarow_filter_size
        )
        global_keys = list(
            map(
                lambda x: x.json["data_row"]["global_key"],
                export_task.get_buffered_stream(),
            )
        )
        assert global_keys.sort() == expected_global_keys.sort()

    def test_batch(
        self,
        configured_batch_project_with_label: Tuple[
            Project, Dataset, DataRow, Label
        ],
        dataset: Dataset,
        image_url: str,
        project_export,
    ):
        project, dataset, *_ = configured_batch_project_with_label
        batch = list(project.batches())[0]
        filters = {
            "last_activity_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
            "label_created_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
            "batch_ids": [batch.uid],
        }
        params = {
            "include_performance_details": True,
            "include_labels": True,
            "media_type_override": MediaType.Image,
        }
        task_name = "TestExportProject:test_batch"
        task = dataset.create_data_rows(
            [
                {"row_data": image_url, "external_id": "my-image"},
            ]
            * 2
        )
        task.wait_till_done()
        data_rows = [result["id"] for result in task.result]
        batch_one = f"batch one {uuid.uuid4()}"

        # This test creates two batches, only one batch should be exporter
        # Creatin second batch that will not be used in the export due to the filter: batch_id
        project.create_batch(batch_one, data_rows)

        export_task = project_export(
            project, task_name, filters=filters, params=params
        )
        assert export_task.has_result()
        assert export_task.has_errors() is False
        assert (
            export_task.get_total_file_size(stream_type=StreamType.RESULT) > 0
        )
        assert (
            export_task.get_total_lines(stream_type=StreamType.RESULT)
            == batch.size
        )



from lbox.exceptions import (
    LabelboxError,
    error_message_for_unparsed_graphql_error,
)

from labelbox.orm.db_object import DbObject
from labelbox.orm.model import Field, Relationship


class ProjectModelConfig(DbObject):
    """A ProjectModelConfig represents an association between a project and a single model config.

    Attributes:
        project_id (str): ID of project to associate
        model_config_id (str): ID of the model configuration
        model_config (ModelConfig): Configuration for model
    """

    project_id = Field.String("project_id", "projectId")
    model_config_id = Field.String("model_config_id", "modelConfigId")
    model_config = Relationship.ToOne("ModelConfig", False, "model_config")

    def delete(self) -> bool:
        """Deletes this association between a model config and this project.

        Returns:
            bool, indicates if the operation was a success.
        """
        query = """mutation DeleteProjectModelConfigPyApi($id: ID!)  {
                    deleteProjectModelConfig(input: {id: $id}) {
                        success
                    }
                }"""

        params = {
            "id": self.uid,
        }

        try:
            result = self.client.execute(query, params)
        except LabelboxError as e:
            if e.message.startswith(
                "Unknown error: "
            ):  # unfortunate hack to handle unparsed graphql errors
                error_content = error_message_for_unparsed_graphql_error(
                    e.message
                )
            else:
                error_content = e.message
            raise LabelboxError(message=error_content) from e

        return result["deleteProjectModelConfig"]["success"]



import pytest

from labelbox.data.annotation_types import (
    Checklist,
    ClassificationAnswer,
    Radio,
    Text,
    ClassificationAnnotation,
)

from pydantic import ValidationError


def test_classification_answer():
    with pytest.raises(ValidationError):
        ClassificationAnswer()

    feature_schema_id = "immunoelectrophoretically"
    name = "my_feature"
    confidence = 0.9
    custom_metrics = [{"name": "metric1", "value": 2}]
    answer = ClassificationAnswer(
        name=name, confidence=confidence, custom_metrics=custom_metrics
    )

    assert answer.feature_schema_id is None
    assert answer.name == name
    assert answer.confidence == confidence
    assert [
        answer.custom_metrics[0].model_dump(exclude_none=True)
    ] == custom_metrics

    answer = ClassificationAnswer(
        feature_schema_id=feature_schema_id, name=name
    )

    assert answer.feature_schema_id == feature_schema_id
    assert answer.name == name


def test_classification():
    answer = "1234"
    classification = ClassificationAnnotation(
        value=Text(answer=answer), name="a classification"
    )
    assert (
        classification.model_dump(exclude_none=True)["value"]["answer"]
        == answer
    )

    with pytest.raises(ValidationError):
        ClassificationAnnotation()


def test_subclass():
    answer = "1234"
    feature_schema_id = "immunoelectrophoretically"
    name = "my_feature"
    with pytest.raises(ValidationError):
        # Should have feature schema info
        classification = ClassificationAnnotation(value=Text(answer=answer))
    classification = ClassificationAnnotation(
        value=Text(answer=answer), name=name
    )
    assert classification.model_dump(exclude_none=True) == {
        "name": name,
        "extra": {},
        "value": {
            "answer": answer,
        },
    }
    classification = ClassificationAnnotation(
        value=Text(answer=answer),
        name=name,
        feature_schema_id=feature_schema_id,
    )
    assert classification.model_dump(exclude_none=True) == {
        "feature_schema_id": feature_schema_id,
        "extra": {},
        "value": {
            "answer": answer,
        },
        "name": name,
    }
    classification = ClassificationAnnotation(
        value=Text(answer=answer),
        feature_schema_id=feature_schema_id,
        name=name,
    )
    assert classification.model_dump(exclude_none=True) == {
        "name": name,
        "feature_schema_id": feature_schema_id,
        "extra": {},
        "value": {
            "answer": answer,
        },
    }


def test_radio():
    answer = ClassificationAnswer(
        name="1",
        confidence=0.81,
        custom_metrics=[{"name": "metric1", "value": 0.99}],
    )
    feature_schema_id = "immunoelectrophoretically"
    name = "my_feature"

    with pytest.raises(ValidationError):
        classification = ClassificationAnnotation(
            value=Radio(answer=answer.name)
        )

    with pytest.raises(ValidationError):
        classification = Radio(answer=[answer])
    classification = Radio(answer=answer)

    assert classification.model_dump(exclude_none=True) == {
        "answer": {
            "name": answer.name,
            "extra": {},
            "confidence": 0.81,
            "custom_metrics": [{"name": "metric1", "value": 0.99}],
        }
    }
    classification = ClassificationAnnotation(
        value=Radio(answer=answer),
        feature_schema_id=feature_schema_id,
        name=name,
        custom_metrics=[{"name": "metric1", "value": 0.99}],
    )
    assert classification.model_dump(exclude_none=True) == {
        "name": name,
        "feature_schema_id": feature_schema_id,
        "extra": {},
        "custom_metrics": [{"name": "metric1", "value": 0.99}],
        "value": {
            "answer": {
                "name": answer.name,
                "extra": {},
                "confidence": 0.81,
                "custom_metrics": [{"name": "metric1", "value": 0.99}],
            },
        },
    }


def test_checklist():
    answer = ClassificationAnswer(
        name="1",
        confidence=0.99,
        custom_metrics=[{"name": "metric1", "value": 2}],
    )
    feature_schema_id = "immunoelectrophoretically"
    name = "my_feature"

    with pytest.raises(ValidationError):
        classification = Checklist(answer=answer.name)

    with pytest.raises(ValidationError):
        classification = Checklist(answer=answer)

    classification = Checklist(answer=[answer])
    assert classification.model_dump(exclude_none=True) == {
        "answer": [
            {
                "name": answer.name,
                "extra": {},
                "confidence": 0.99,
                "custom_metrics": [{"name": "metric1", "value": 2}],
            }
        ]
    }
    classification = ClassificationAnnotation(
        value=Checklist(answer=[answer]),
        feature_schema_id=feature_schema_id,
        name=name,
    )
    assert classification.model_dump(exclude_none=True) == {
        "name": name,
        "feature_schema_id": feature_schema_id,
        "extra": {},
        "value": {
            "answer": [
                {
                    "name": answer.name,
                    "extra": {},
                    "confidence": 0.99,
                    "custom_metrics": [{"name": "metric1", "value": 2}],
                }
            ]
        },
    }



from labelbox.schema.data_row_metadata import _DeleteBatchDataRowMetadata
from labelbox.schema.identifiable import GlobalKey, UniqueId


def test_dict_delete_data_row_batch():
    obj = _DeleteBatchDataRowMetadata(
        data_row_identifier=UniqueId("abcd"),
        schema_ids=["clqh77tyk000008l2a9mjesa1", "clqh784br000008jy0yuq04fy"],
    )
    assert obj.model_dump() == {
        "data_row_identifier": {"id": "abcd", "id_type": "ID"},
        "schema_ids": [
            "clqh77tyk000008l2a9mjesa1",
            "clqh784br000008jy0yuq04fy",
        ],
    }

    obj = _DeleteBatchDataRowMetadata(
        data_row_identifier=GlobalKey("fegh"),
        schema_ids=["clqh77tyk000008l2a9mjesa1", "clqh784br000008jy0yuq04fy"],
    )
    assert obj.model_dump() == {
        "data_row_identifier": {"id": "fegh", "id_type": "GKEY"},
        "schema_ids": [
            "clqh77tyk000008l2a9mjesa1",
            "clqh784br000008jy0yuq04fy",
        ],
    }


def test_dict_delete_data_row_batch_by_alias():
    obj = _DeleteBatchDataRowMetadata(
        data_row_identifier=UniqueId("abcd"),
        schema_ids=["clqh77tyk000008l2a9mjesa1", "clqh784br000008jy0yuq04fy"],
    )
    assert obj.model_dump(by_alias=True) == {
        "dataRowIdentifier": {"id": "abcd", "idType": "ID"},
        "schemaIds": ["clqh77tyk000008l2a9mjesa1", "clqh784br000008jy0yuq04fy"],
    }

    obj = _DeleteBatchDataRowMetadata(
        data_row_identifier=GlobalKey("fegh"),
        schema_ids=["clqh77tyk000008l2a9mjesa1", "clqh784br000008jy0yuq04fy"],
    )
    assert obj.model_dump(by_alias=True) == {
        "dataRowIdentifier": {"id": "fegh", "idType": "GKEY"},
        "schemaIds": ["clqh77tyk000008l2a9mjesa1", "clqh784br000008jy0yuq04fy"],
    }



import time
from datetime import datetime, timedelta, timezone

import pytest


def simple_ontology():
    classifications = [
        {
            "name": "test_ontology",
            "instructions": "Which class is this?",
            "type": "radio",
            "options": [
                {"value": c, "label": c} for c in ["one", "two", "three"]
            ],
            "required": True,
        }
    ]

    return {"tools": [], "classifications": classifications}


def test_project_editor_setup(client, project, rand_gen):
    ontology_name = f"test_project_editor_setup_ontology_name-{rand_gen(str)}"
    ontology = client.create_ontology(ontology_name, simple_ontology())
    now = datetime.now().astimezone(timezone.utc)
    project.connect_ontology(ontology)
    assert now - project.setup_complete <= timedelta(seconds=3)
    assert now - project.last_activity_time <= timedelta(seconds=3)
    assert project.labeling_frontend().name == "Editor"
    assert project.organization() == client.get_organization()
    assert project.created_by() == client.get_user()
    assert project.ontology().name == ontology_name
    # Make sure that setup only creates one ontology
    time.sleep(3)  # Search takes a second
    assert [
        ontology.name for ontology in client.get_ontologies(ontology_name)
    ] == [ontology_name]


def test_project_connect_ontology_cant_call_multiple_times(
    client, project, rand_gen
):
    ontology_name = f"test_project_editor_setup_ontology_name-{rand_gen(str)}"
    ontology = client.create_ontology(ontology_name, simple_ontology())
    project.connect_ontology(ontology)
    with pytest.raises(ValueError):
        project.connect_ontology(ontology)



from typing import Optional, Union, Tuple
from enum import Enum

import cv2
import geojson
import numpy as np

from shapely.geometry import Polygon as SPolygon

from .geometry import Geometry
from .point import Point


class Rectangle(Geometry):
    """Represents a 2d rectangle. Also known as a bounding box

    >>> Rectangle(start=Point(x=0, y=0), end=Point(x=1, y=1))

    Args:
        start (Point): Top left coordinate of the rectangle
        end (Point): Bottom right coordinate of the rectangle
    """

    start: Point
    end: Point

    @property
    def geometry(self) -> geojson.geometry.Geometry:
        return geojson.Polygon(
            [
                [
                    [self.start.x, self.start.y],
                    [self.start.x, self.end.y],
                    [self.end.x, self.end.y],
                    [self.end.x, self.start.y],
                    [self.start.x, self.start.y],
                ]
            ]
        )

    @classmethod
    def from_shapely(cls, shapely_obj: SPolygon) -> "Rectangle":
        """Transforms a shapely object.

        If the provided shape is a non-rectangular polygon, a rectangle will be
        returned based on the min and max x,y values."""
        if not isinstance(shapely_obj, SPolygon):
            raise TypeError(
                f"Expected Shapely Polygon. Got {shapely_obj.geom_type}"
            )

        min_x, min_y, max_x, max_y = shapely_obj.bounds

        start = [min_x, min_y]
        end = [max_x, max_y]

        return Rectangle(
            start=Point(x=start[0], y=start[1]), end=Point(x=end[0], y=end[1])
        )

    def draw(
        self,
        height: Optional[int] = None,
        width: Optional[int] = None,
        canvas: Optional[np.ndarray] = None,
        color: Union[int, Tuple[int, int, int]] = (255, 255, 255),
        thickness: int = -1,
    ) -> np.ndarray:
        """
        Draw the rectangle onto a 3d mask
        Args:
            height (int): height of the mask
            width (int): width of the mask
            color (int): color for the polygon.
                  RGB values by default but if a 2D canvas is provided this can set this to an int.
            thickness (int): How thick to make the rectangle border. -1 fills in the rectangle
            canvas (np.ndarray): Canvas to draw rectangle on
        Returns:
            numpy array representing the mask with the rectangle drawn on it.
        """
        canvas = self.get_or_create_canvas(height, width, canvas)
        pts = np.array(self.geometry["coordinates"]).astype(np.int32)
        if thickness == -1:
            return cv2.fillPoly(canvas, pts, color)
        return cv2.polylines(canvas, pts, True, color, thickness)

    @classmethod
    def from_xyhw(cls, x: float, y: float, h: float, w: float) -> "Rectangle":
        """Create Rectangle from x,y, height width format"""
        return cls(start=Point(x=x, y=y), end=Point(x=x + w, y=y + h))


class RectangleUnit(Enum):
    INCHES = "INCHES"
    PIXELS = "PIXELS"
    POINTS = "POINTS"
    PERCENT = "PERCENT"


class DocumentRectangle(Rectangle):
    """Represents a 2d rectangle on a Document

    >>> Rectangle(
    >>>     start=Point(x=0, y=0),
    >>>     end=Point(x=1, y=1),
    >>>     page=4,
    >>>     unit=RectangleUnits.POINTS
    >>> )

    Args:
        start (Point): Top left coordinate of the rectangle
        end (Point): Bottom right coordinate of the rectangle
        page (int): Page number of the document
        unit (RectangleUnits): Units of the rectangle
    """

    page: int
    unit: RectangleUnit



from labelbox.data.annotation_types.base_annotation import BaseAnnotation
from labelbox.data.mixins import ConfidenceMixin, CustomMetricsMixin
from pydantic import BaseModel


class PromptText(ConfidenceMixin, CustomMetricsMixin, BaseModel):
    """Prompt text for LLM data generation

    >>> PromptText(answer = "some text answer",
    >>>    confidence = 0.5,
    >>>    custom_metrics = [
    >>>     {
    >>>         "name": "iou",
    >>>         "value": 0.1
    >>>     }])
    """

    answer: str


class PromptClassificationAnnotation(
    BaseAnnotation, ConfidenceMixin, CustomMetricsMixin
):
    """Prompt annotation (non localized)

    >>> PromptClassificationAnnotation(
    >>>     value=PromptText(answer="my caption message"),
    >>>     feature_schema_id="my-feature-schema-id"
    >>> )

    Args:
        name (Optional[str])
        feature_schema_id (Optional[Cuid])
        value (Union[Text])
    """

    value: PromptText



from typing import Union

from lbox import exceptions  # type: ignore

from labelbox.schema.foundry.app import APP_FIELD_NAMES, App
from labelbox.schema.identifiables import DataRowIds, GlobalKeys, IdType
from labelbox.schema.task import Task


class FoundryClient:
    def __init__(self, client):
        self.client = client

    def _create_app(self, app: App) -> App:
        field_names_str = "\n".join(APP_FIELD_NAMES)
        query_str = f"""
            mutation CreateFoundryAppPyApi(
                $name: String!, $modelId: ID!, $ontologyId: ID!, $description: String, $inferenceParams: Json!, $classToSchemaId: Json!
            ){{
                createModelFoundryApp(input: {{
                    name: $name 
                    modelId: $modelId 
                    ontologyId: $ontologyId
                    description: $description
                    inferenceParams: $inferenceParams
                    classToSchemaId: $classToSchemaId
                }})
                {{
                    {field_names_str}
                }}
            }}
            """

        params = app.model_dump(by_alias=True, exclude={"id"})

        try:
            response = self.client.execute(query_str, params)
        except exceptions.LabelboxError as e:
            raise exceptions.LabelboxError("Unable to create app", e)
        return App(**response["createModelFoundryApp"])

    def _get_app(self, id: str) -> App:
        field_names_str = "\n".join(APP_FIELD_NAMES)

        query_str = f"""
            query GetFoundryAppByIdPyApi($id: ID!) {{
                findModelFoundryApp(where: {{id: $id}}) {{
                    {field_names_str}
                }}
            }}
        """
        params = {"id": id}

        try:
            response = self.client.execute(query_str, params)
        except exceptions.InvalidQueryError:
            raise exceptions.ResourceNotFoundError(App, params)
        except Exception as e:
            raise exceptions.LabelboxError(f"Unable to get app with id {id}", e)
        return App(**response["findModelFoundryApp"])

    def _delete_app(self, id: str) -> None:
        query_str = """
            mutation DeleteFoundryAppPyApi($id: ID!) {
                deleteModelFoundryApp(id: $id) {
                    success
                }
            }
        """
        params = {"id": id}
        try:
            self.client.execute(query_str, params)
        except Exception as e:
            raise exceptions.LabelboxError(
                f"Unable to delete app with id {id}", e
            )

    def run_app(
        self,
        model_run_name: str,
        data_rows: Union[DataRowIds, GlobalKeys],
        app_id: str,
    ) -> Task:
        app = self._get_app(app_id)

        params = {
            "modelId": str(app.model_id),
            "name": model_run_name,
            "classToSchemaId": app.class_to_schema_id,
            "inferenceParams": app.inference_params,
            "ontologyId": app.ontology_id,
        }

        data_rows_key = (
            "dataRowIds"
            if data_rows.id_type == IdType.DataRowId
            else "globalKeys"
        )
        params[data_rows_key] = list(data_rows)

        query = """
        mutation CreateModelJobPyApi($input: CreateModelJobForDataRowsInput!) {
            createModelJobForDataRows(input: $input) {
                taskId
                __typename
            }
        }
        """
        try:
            response = self.client.execute(query, {"input": params})
        except Exception as e:
            raise exceptions.LabelboxError("Unable to run foundry app", e)
        task_id = response["createModelJobForDataRows"]["taskId"]
        return Task.get_task(self.client, task_id)



from typing import Dict, List, Optional

from pydantic import BaseModel, Field

from labelbox.schema.data_row import DataRowMetadataField


class ModelEvalutationTemplateRowData(BaseModel):
    type: str = Field(
        default="application/vnd.labelbox.conversational.model-chat-evaluation",
        frozen=True,
    )
    draft: bool = Field(default=True, frozen=True)
    rootMessageIds: List[str] = Field(default=[])
    actors: Dict = Field(default={})
    version: int = Field(default=2, frozen=True)
    messages: Dict = Field(default={})


class ModelEvaluationTemplate(BaseModel):
    """
    Use this class to create a model evaluation data row.

    Examples:
        >>> data = ModelEvaluationTemplate()
        >>> data.row_data.rootMessageIds = ["root1"]
        >>> vector = [random.uniform(1.0, 2.0) for _ in range(embedding.dims)]
        >>> data.embeddings = [...]
        >>> data.metadata_fields = [...]
        >>> data.attachments = [...]
        >>> content = data.model_dump()
        >>> task = dataset.create_data_rows([content])
    """

    row_data: ModelEvalutationTemplateRowData = Field(
        default=ModelEvalutationTemplateRowData()
    )
    global_key: Optional[str] = None
    attachments: List[Dict] = Field(default=[])
    embeddings: List[Dict] = Field(default=[])
    metadata_fields: List[DataRowMetadataField] = Field(default=[])



from typing import Optional

import pytest
from faker import Faker

from labelbox import ProjectRole
from labelbox.schema.invite import Invite

faker = Faker()


def find_invite_and_id_by_email(
    queries, client, invite_email
) -> Optional[Invite]:
    """For security reasons, invite uid is not returned for a query for a single invite.
    This function is a workaround to find the invite uid by email using another query that returns all invites.
    """
    outstanding_invites = queries.get_invites(client)

    for outstanding_invite in outstanding_invites:
        if outstanding_invite.email == invite_email:
            return outstanding_invite
    return None


@pytest.fixture
def org_invite(client, organization, environ, queries):
    role = client.get_roles()["LABELER"]

    dummy_email = "none+{}@labelbox.com".format(
        "".join(faker.random_letters(26))
    )
    invite_limit = organization.invite_limit()

    if environ.value == "prod":
        assert (
            invite_limit.remaining > 0
        ), "No invites available for the account associated with this key."
    elif environ.value != "staging":
        # Cannot run against local
        return

    invite = organization.invite_user(dummy_email, role)

    yield invite, invite_limit

    found_invite = find_invite_and_id_by_email(queries, client, invite.email)
    assert found_invite is not None, "Invite not found"
    queries.cancel_invite(client, found_invite.uid)


@pytest.fixture
def project_role_1(client, project_pack):
    project_1, _ = project_pack
    roles = client.get_roles()
    return ProjectRole(project=project_1, role=roles["LABELER"])


@pytest.fixture
def project_role_2(client, project_pack):
    _, project_2 = project_pack
    roles = client.get_roles()
    return ProjectRole(project=project_2, role=roles["REVIEWER"])


@pytest.fixture
def create_project_invite(
    client, organization, project_pack, queries, project_role_1, project_role_2
):
    roles = client.get_roles()
    dummy_email = "none+{}@labelbox.com".format(
        "".join(faker.random_letters(26))
    )
    invite = organization.invite_user(
        dummy_email,
        roles["NONE"],
        project_roles=[project_role_1, project_role_2],
    )

    yield invite

    found_invite = find_invite_and_id_by_email(queries, client, invite.email)
    assert found_invite is not None, "Invite not found"
    queries.cancel_invite(client, found_invite.uid)


def test_org_invite(client, organization, environ, queries, org_invite):
    invite, invite_limit = org_invite
    role = client.get_roles()["LABELER"]

    assert (
        invite.uid == "invited"
    )  # for security reasons we don't return the invite uid

    if environ.value == "prod":
        invite_limit_after = organization.invite_limit()
        # One user added
        assert invite_limit.remaining - invite_limit_after.remaining == 1
        # An invite shouldn't effect the user count until after it is accepted

    found_invite = find_invite_and_id_by_email(queries, client, invite.email)

    assert found_invite is not None, "Invite not found"
    org_role = found_invite.organization_role_name.lower()
    assert (
        org_role == role.name.lower()
    ), "Role should be labeler. Found {org_role} "


def test_cancel_invite(
    client,
    organization,
    queries,
):
    role = client.get_roles()["LABELER"]
    dummy_email = "none+{}@labelbox.com".format(
        "".join(faker.random_letters(26))
    )
    invite = organization.invite_user(dummy_email, role)
    found_invite = find_invite_and_id_by_email(queries, client, invite.email)
    assert found_invite is not None, "Invite not found"
    invite.uid = found_invite.uid  # for security reasons we don't return the invite uid directly but need it in order to cancel the invite

    queries.cancel_invite(client, invite.uid)
    outstanding_invites = [i.uid for i in queries.get_invites(client)]
    assert invite.uid not in outstanding_invites


def test_project_invite(
    client,
    organization,
    project_pack,
    queries,
    create_project_invite,
    project_role_1,
    project_role_2,
):
    create_project_invite
    project_1, _ = project_pack
    roles = client.get_roles()

    project_invite = next(queries.get_project_invites(client, project_1.uid))
    assert set(
        [
            (proj_invite.project.uid, proj_invite.role.uid)
            for proj_invite in project_invite.project_roles
        ]
    ) == set(
        [
            (proj_role.project.uid, proj_role.role.uid)
            for proj_role in [project_role_1, project_role_2]
        ]
    )

    assert set(
        [
            (proj_invite.project.uid, proj_invite.role.uid)
            for proj_invite in project_invite.project_roles
        ]
    ) == set(
        [
            (proj_role.project.uid, proj_role.role.uid)
            for proj_role in [project_role_1, project_role_2]
        ]
    )

    project_members = project_1.members()

    project_member = [
        member
        for member in project_members
        if member.user().uid == client.get_user().uid
    ]

    assert len(project_member) == 1
    project_member = project_member[0]

    assert project_member.access_from == "ORGANIZATION"
    assert project_member.role().name.upper() == roles["ADMIN"].name.upper()


@pytest.mark.skip(
    "Unable to programatically create user without accepting an email invite. Add back once there is a workaround."
)
def test_member_management(client, organization, project, project_based_user):
    roles = client.get_roles()
    assert not len(list(project_based_user.projects()))
    for role in [roles["LABELER"], roles["REVIEWER"]]:
        project_based_user.upsert_project_role(project, role=role)
        members = project.members()
        is_member = False
        for member in members:
            if member.user().uid == project_based_user.uid:
                is_member = True
                assert member.role().name.upper() == role.name.upper()
                break
        assert is_member

    project_based_user.remove_from_project(project)
    is_member = False
    for member in project.members():
        assert member.user().uid != project_based_user.uid

    assert (
        project_based_user.org_role().name.upper() == roles["NONE"].name.upper()
    )
    for role in [
        roles["TEAM_MANAGER"],
        roles["ADMIN"],
        roles["LABELER"],
        roles["REVIEWER"],
    ]:
        project_based_user.update_org_role(role)
        project_based_user.org_role().name.upper() == role.name.upper()

    organization.remove_user(project_based_user)
    for user in organization.users():
        assert project_based_user.uid != user.uid



from datetime import datetime
from string import Template
from typing import Any, Dict, List, Optional, Union

from lbox.exceptions import ResourceNotFoundError
from pydantic import BaseModel, Field, model_serializer, model_validator

from labelbox.pagination import PaginatedCollection
from labelbox.schema.labeling_service_status import LabelingServiceStatus
from labelbox.schema.media_type import MediaType
from labelbox.schema.search_filters import SearchFilter, build_search_filter
from labelbox.utils import _CamelCaseMixin, sentence_case

from .ontology_kind import EditorTaskType

GRAPHQL_QUERY_SELECTIONS = """
                id
                name
                boostRequestedAt
                boostUpdatedAt
                boostRequestedBy
                boostStatus
                dataRowsCount
                dataRowsDoneCount
                dataRowsInReviewCount
                dataRowsInReworkCount
                tasksTotalCount
                tasksCompletedCount
                tasksRemainingCount
                mediaType
                editorTaskType
                tags {
                    id
                    text
                    color
                    type
                }
            """


class LabelingServiceDashboardTags(BaseModel):
    id: str
    text: str
    color: str
    type: str


class LabelingServiceDashboard(_CamelCaseMixin):
    """
    Represent labeling service data for a project

    NOTE on tasks vs data rows. A task is a unit of work that is assigned to a user. A data row is a unit of data that needs to be labeled.
        In the current implementation a task represents a single data row. However tasks only exists when a labeler start labeling a data row.
        So if a data row is not labeled, it will not have a task associated with it. Therefore the number of tasks can be less than the number of data rows.

    Attributes:
        id (str): project id
        name (str): project name
        status (LabelingServiceStatus): status of the labeling service
        data_rows_count (int): total number of data rows batched in the project
        tasks_completed_count (int): number of tasks completed (in the Done queue)
        tasks_remaining_count (int): number of tasks remaining (i.e. tasks in progress), None if labeling has not started
        tags (List[LabelingServiceDashboardTags]): tags associated with the project
        media_type (MediaType): media type of the project
        editor_task_type (EditorTaskType): editor task type of the project
        client (Any): labelbox client
    """

    id: str = Field(frozen=True)
    name: str = Field(frozen=True)
    created_at: Optional[datetime] = Field(frozen=True, default=None)
    updated_at: Optional[datetime] = Field(frozen=True, default=None)
    created_by_id: Optional[str] = Field(frozen=True, default=None)
    status: Optional[LabelingServiceStatus] = Field(frozen=True, default=None)
    data_rows_count: int = Field(frozen=True)
    tasks_completed_count: int = Field(frozen=True)
    tasks_remaining_count: Optional[int] = Field(frozen=True, default=None)
    media_type: Optional[MediaType] = Field(frozen=True, default=None)
    editor_task_type: Optional[EditorTaskType] = Field(
        frozen=True, default=None
    )
    tags: Optional[List[LabelingServiceDashboardTags]] = Field(
        frozen=True, default=None
    )

    client: Any  # type Any to avoid circular import from client

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def service_type(self):
        """
        Descriptive labeling service definition by media type and editor task type
        """
        if self.media_type is None:
            return None

        if self.editor_task_type is None:
            return sentence_case(self.media_type.value)

        if (
            self.editor_task_type == EditorTaskType.OfflineModelChatEvaluation
            and self.media_type == MediaType.Conversational
        ):
            return "Offline chat evaluation"

        if (
            self.editor_task_type == EditorTaskType.ModelChatEvaluation
            and self.media_type == MediaType.Conversational
        ):
            return "Live chat evaluation"

        if (
            self.editor_task_type == EditorTaskType.ResponseCreation
            and self.media_type == MediaType.Text
        ):
            return "Response creation"

        if (
            self.media_type == MediaType.LLMPromptCreation
            or self.media_type == MediaType.LLMPromptResponseCreation
        ):
            return "Prompt response creation"

        return sentence_case(self.media_type.value)

    @classmethod
    def get(cls, client, project_id: str) -> "LabelingServiceDashboard":
        """
        Returns the labeling service associated with the project.

        Raises:
            ResourceNotFoundError: If the project does not have a labeling service.
        """
        query = f"""
                    query GetProjectByIdPyApi($id: ID!) {{
                    getProjectById(input: {{id: $id}}) {{
                        {GRAPHQL_QUERY_SELECTIONS}
                        }}
                    }}
                """
        result = client.execute(query, {"id": project_id})
        if result["getProjectById"] is None:
            raise ResourceNotFoundError(
                message="The project does not have a labeling service data yet."
            )
        data = result["getProjectById"]
        data["client"] = client
        return cls(**data)

    @classmethod
    def get_all(
        cls,
        client,
        search_query: Optional[List[SearchFilter]] = None,
    ) -> PaginatedCollection:
        if search_query is not None:
            template = Template(
                """query SearchProjectsPyApi($$first: Int, $$from: String) {
                            searchProjects(input: {after: $$from, searchQuery: $search_query, size: $$first})
                                {
                                    nodes { $labeling_dashboard_selections }
                                    pageInfo { endCursor }
                                }
                            }
                        """
            )
        else:
            template = Template(
                """query SearchProjectsPyApi($$first: Int, $$from: String) {
                            searchProjects(input: {after: $$from, size: $$first})
                                {
                                    nodes { $labeling_dashboard_selections }
                                    pageInfo { endCursor }
                                }
                            }
                        """
            )
        query_str = template.substitute(
            labeling_dashboard_selections=GRAPHQL_QUERY_SELECTIONS,
            search_query=build_search_filter(search_query)
            if search_query
            else None,
        )
        params: Dict[str, Union[str, int]] = {}

        def convert_to_labeling_service_dashboard(client, data):
            data["client"] = client
            return LabelingServiceDashboard(**data)

        return PaginatedCollection(
            client=client,
            query=query_str,
            params=params,
            dereferencing=["searchProjects", "nodes"],
            obj_class=convert_to_labeling_service_dashboard,
            cursor_path=["searchProjects", "pageInfo", "endCursor"],
        )

    @model_validator(mode="before")
    def convert_boost_data(cls, data):
        if "boostStatus" in data:
            data["status"] = LabelingServiceStatus(data.pop("boostStatus"))

        if "boostRequestedAt" in data:
            data["created_at"] = data.pop("boostRequestedAt")

        if "boostUpdatedAt" in data:
            data["updated_at"] = data.pop("boostUpdatedAt")

        if "boostRequestedBy" in data:
            data["created_by_id"] = data.pop("boostRequestedBy")

        tasks_remaining_count = data.get("tasksRemainingCount", 0)
        tasks_total_count = data.get("tasksTotalCount", 0)
        # to avoid confusion, setting tasks_completed_count to None if none of tasks has even completed an none are in flight
        if tasks_total_count == 0 and tasks_remaining_count == 0:
            data.pop("tasksRemainingCount")

        return data

    @model_serializer(mode="wrap")
    def ser_model(self, handler):
        row = handler(self)
        row.pop("client")
        row["service_type"] = self.service_type
        return row



from typing import Optional
from uuid import uuid4

from labelbox.utils import _CamelCaseMixin, is_exactly_one_set
from pydantic import model_validator, Field
from uuid import uuid4

from ....annotated_types import Cuid


class DataRow(_CamelCaseMixin):
    id: Optional[str] = None
    global_key: Optional[str] = None

    @model_validator(mode="after")
    def must_set_one(self):
        if not is_exactly_one_set(self.id, self.global_key):
            raise ValueError("Must set either id or global_key")
        return self


class NDJsonBase(_CamelCaseMixin):
    uuid: Optional[str] = Field(default_factory=lambda: str(uuid4()))
    data_row: DataRow


class NDAnnotation(NDJsonBase):
    name: Optional[str] = None
    schema_id: Optional[Cuid] = None
    message_id: Optional[str] = None
    page: Optional[int] = None
    unit: Optional[str] = None

    @model_validator(mode="after")
    def must_set_one(self):
        if (not hasattr(self, "schema_id") or self.schema_id is None) and (
            not hasattr(self, "name") or self.name is None
        ):
            raise ValueError("Schema id or name are not set. Set either one.")
        return self
